import imp

from django.conf import settings
from django.utils.importlib import import_module

from avocado.exceptions import RegisterError, AlreadyRegisteredError
from avocado.utils.camel import uncamel

class Library(object):
    superclass = None
    module_name = None
    suffix = ''
    default = None

    def __init__(self):
        self._cache = {}
        self._loaded = False
        self._discovery = False

    def _clean_name(self, name):
        if self.suffix and name.endswith(self.suffix):
            name = name[:-len(self.suffix)]
        return uncamel(name)

    def _get_class_name(self, klass):
        if hasattr(klass, 'name'):
            return klass.name
        return self._clean_name(klass.__name__)

    def add_object(self, class_name, obj, errmsg=''):
        store = self.get_store()
        # if already registered under the same name, test if the classes are
        # the same -- replace if true, throw error if not
        if store.has_key(class_name):
            robj = store.get(class_name)
            if obj.__class__ is not robj.__class__:
                raise AlreadyRegisteredError, errmsg
        store.update({class_name: obj})

    def remove_object(self, class_name):
        store = self.get_store()
        if store.has_key(class_name):
            store.remove(class_name)

    def get_store(self):
        if not self._cache and not self._loaded:
            self.autodiscover()
        return self._cache

    def register_object(self, class_name, obj):
        self.add_object(class_name, obj)

    def unregister_object(self, class_name):
        self.remove_object(class_name)

    def register(self, klass):
        if not issubclass(klass, self.superclass):
            raise RegisterError, '%s must be a subclass of %s' % (repr(klass),
                self.superclass.__name__)

        obj = klass()
        class_name = self._get_class_name(klass)

        self.register_object(class_name, obj)
        return klass

    def unregister(self, klass):
        class_name = self._get_class_name(klass)
        self.unregister_object(class_name)

    def choices(self):
        "Returns a list of tuples that can be used as choices in a form."
        store = self.get_store()
        choices = []
        for name in store.keys():
            choices.append((name, name))
        return choices

    def get(self, name):
        "Retrieve the cached instance given the name it is registered under."
        return self.get_store().get(name, self.default)

    def autodiscover(self):
        if self._discovery:
            return

        self._discovery = True

        for app in settings.INSTALLED_APPS:
            try:
                app_path = import_module(app).__path__
            except AttributeError:
                continue
            try:
                imp.find_module(self.module_name, app_path)
            except ImportError:
                continue

            import_module('%s.%s' % (app, self.module_name))

        self._discovery = False



# encoding: utf-8
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models

class Migration(SchemaMigration):

    def forwards(self, orm):
        
        # Adding field 'Perspective.created'
        db.add_column('avocado_perspective', 'created', self.gf('django.db.models.fields.DateTimeField')(default=datetime.datetime(2011, 8, 9, 17, 36, 9, 293249)), keep_default=False)

        # Adding field 'Perspective.modified'
        db.add_column('avocado_perspective', 'modified', self.gf('django.db.models.fields.DateTimeField')(default=datetime.datetime(2011, 8, 9, 17, 36, 17, 501402)), keep_default=False)

        # Changing field 'Perspective.store'
        db.alter_column('avocado_perspective', 'store', self.gf('avocado.store.fields.JSONField')(editable=False))

        # Adding field 'Report.created'
        db.add_column('avocado_report', 'created', self.gf('django.db.models.fields.DateTimeField')(default=datetime.datetime(2011, 8, 9, 17, 36, 22, 805747)), keep_default=False)

        # Adding field 'Report.modified'
        db.add_column('avocado_report', 'modified', self.gf('django.db.models.fields.DateTimeField')(default=datetime.datetime(2011, 8, 9, 17, 36, 32, 445973)), keep_default=False)

        # Adding field 'Scope.created'
        db.add_column('avocado_scope', 'created', self.gf('django.db.models.fields.DateTimeField')(default=datetime.datetime(2011, 8, 9, 17, 36, 42, 598399)), keep_default=False)

        # Adding field 'Scope.modified'
        db.add_column('avocado_scope', 'modified', self.gf('django.db.models.fields.DateTimeField')(default=datetime.datetime(2011, 8, 9, 17, 36, 46, 950524)), keep_default=False)

        # Changing field 'Scope.store'
        db.alter_column('avocado_scope', 'store', self.gf('avocado.store.fields.JSONField')(editable=False))


    def backwards(self, orm):
        
        # Deleting field 'Perspective.created'
        db.delete_column('avocado_perspective', 'created')

        # Deleting field 'Perspective.modified'
        db.delete_column('avocado_perspective', 'modified')

        # Changing field 'Perspective.store'
        db.alter_column('avocado_perspective', 'store', self.gf('avocado.store.fields.JSONField')(null=True, editable=False))

        # Deleting field 'Report.created'
        db.delete_column('avocado_report', 'created')

        # Deleting field 'Report.modified'
        db.delete_column('avocado_report', 'modified')

        # Deleting field 'Scope.created'
        db.delete_column('avocado_scope', 'created')

        # Deleting field 'Scope.modified'
        db.delete_column('avocado_scope', 'modified')

        # Changing field 'Scope.store'
        db.alter_column('avocado_scope', 'store', self.gf('avocado.store.fields.JSONField')(null=True, editable=False))


    models = {
        'auth.group': {
            'Meta': {'object_name': 'Group'},
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '80'}),
            'permissions': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Permission']", 'symmetrical': 'False', 'blank': 'True'})
        },
        'auth.permission': {
            'Meta': {'ordering': "('content_type__app_label', 'content_type__model', 'codename')", 'unique_together': "(('content_type', 'codename'),)", 'object_name': 'Permission'},
            'codename': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['contenttypes.ContentType']"}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '50'})
        },
        'auth.user': {
            'Meta': {'object_name': 'User'},
            'date_joined': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'email': ('django.db.models.fields.EmailField', [], {'max_length': '75', 'blank': 'True'}),
            'first_name': ('django.db.models.fields.CharField', [], {'max_length': '30', 'blank': 'True'}),
            'groups': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Group']", 'symmetrical': 'False', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_active': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'is_staff': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'is_superuser': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'last_login': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'last_name': ('django.db.models.fields.CharField', [], {'max_length': '30', 'blank': 'True'}),
            'password': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'user_permissions': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Permission']", 'symmetrical': 'False', 'blank': 'True'}),
            'username': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '30'})
        },
        'avocado.category': {
            'Meta': {'ordering': "('order', 'name')", 'object_name': 'Category'},
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'order': ('django.db.models.fields.FloatField', [], {'default': '0'}),
            'parent': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'children'", 'null': 'True', 'to': "orm['avocado.Category']"})
        },
        'avocado.column': {
            'Meta': {'ordering': "('name',)", 'object_name': 'Column'},
            'category': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['avocado.Category']", 'null': 'True', 'blank': 'True'}),
            'csv_fmtr': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'description': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'fields': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['avocado.Field']", 'through': "orm['avocado.ColumnField']", 'symmetrical': 'False'}),
            'html_fmtr': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_public': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'keywords': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'note': ('django.db.models.fields.TextField', [], {'null': 'True'}),
            'order': ('django.db.models.fields.FloatField', [], {'default': '0'}),
            'reviewed': ('django.db.models.fields.DateTimeField', [], {}),
            'search_doc': ('django.db.models.fields.TextField', [], {'null': 'True'}),
            'status': ('django.db.models.fields.CharField', [], {'max_length': '40'})
        },
        'avocado.columnfield': {
            'Meta': {'ordering': "('order',)", 'object_name': 'ColumnField'},
            'concept': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'conceptfields'", 'to': "orm['avocado.Column']"}),
            'field': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['avocado.Field']"}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'order': ('django.db.models.fields.FloatField', [], {'default': '0'})
        },
        'avocado.criterion': {
            'Meta': {'ordering': "('name',)", 'object_name': 'Criterion'},
            'category': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['avocado.Category']", 'null': 'True', 'blank': 'True'}),
            'description': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'fields': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['avocado.Field']", 'through': "orm['avocado.CriterionField']", 'symmetrical': 'False'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_public': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'keywords': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'note': ('django.db.models.fields.TextField', [], {'null': 'True'}),
            'order': ('django.db.models.fields.FloatField', [], {'default': '0'}),
            'reviewed': ('django.db.models.fields.DateTimeField', [], {}),
            'search_doc': ('django.db.models.fields.TextField', [], {'null': 'True'}),
            'status': ('django.db.models.fields.CharField', [], {'max_length': '40'}),
            'viewset': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        'avocado.criterionfield': {
            'Meta': {'ordering': "('order',)", 'object_name': 'CriterionField'},
            'concept': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'conceptfields'", 'to': "orm['avocado.Criterion']"}),
            'field': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['avocado.Field']"}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'order': ('django.db.models.fields.FloatField', [], {'default': '0'}),
            'required': ('django.db.models.fields.BooleanField', [], {'default': 'True'})
        },
        'avocado.field': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_name', 'model_name', 'field_name'),)", 'object_name': 'Field'},
            'app_name': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'chart_title': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'chart_xaxis': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'chart_yaxis': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'choices_handler': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'description': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'enable_choices': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'field_name': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'group': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['auth.Group']", 'null': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_public': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'keywords': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'model_name': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'note': ('django.db.models.fields.TextField', [], {'null': 'True'}),
            'reviewed': ('django.db.models.fields.DateTimeField', [], {'null': 'True'}),
            'search_doc': ('django.db.models.fields.TextField', [], {'null': 'True', 'db_index': 'True'}),
            'sites': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['sites.Site']", 'symmetrical': 'False', 'blank': 'True'}),
            'status': ('django.db.models.fields.CharField', [], {'max_length': '40', 'null': 'True', 'blank': 'True'}),
            'translator': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'})
        },
        'avocado.perspective': {
            'Meta': {'object_name': 'Perspective'},
            'created': ('django.db.models.fields.DateTimeField', [], {}),
            'definition': ('django.db.models.fields.TextField', [], {'null': 'True'}),
            'description': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'keywords': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'modified': ('django.db.models.fields.DateTimeField', [], {}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'store': ('avocado.store.fields.JSONField', [], {'editable': 'False'}),
            'timestamp': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime(2011, 8, 9, 17, 35, 47, 433504)'}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['auth.User']", 'null': 'True', 'blank': 'True'})
        },
        'avocado.report': {
            'Meta': {'object_name': 'Report'},
            'created': ('django.db.models.fields.DateTimeField', [], {}),
            'description': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'keywords': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'modified': ('django.db.models.fields.DateTimeField', [], {}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'perspective': ('django.db.models.fields.related.OneToOneField', [], {'to': "orm['avocado.Perspective']", 'unique': 'True'}),
            'scope': ('django.db.models.fields.related.OneToOneField', [], {'to': "orm['avocado.Scope']", 'unique': 'True'}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['auth.User']", 'null': 'True', 'blank': 'True'})
        },
        'avocado.scope': {
            'Meta': {'object_name': 'Scope'},
            'cnt': ('django.db.models.fields.PositiveIntegerField', [], {}),
            'created': ('django.db.models.fields.DateTimeField', [], {}),
            'definition': ('django.db.models.fields.TextField', [], {'null': 'True'}),
            'description': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'keywords': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'modified': ('django.db.models.fields.DateTimeField', [], {}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'store': ('avocado.store.fields.JSONField', [], {'editable': 'False'}),
            'timestamp': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime(2011, 8, 9, 17, 35, 47, 433504)'}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['auth.User']", 'null': 'True', 'blank': 'True'})
        },
        'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        'sites.site': {
            'Meta': {'ordering': "('domain',)", 'object_name': 'Site', 'db_table': "'django_site'"},
            'domain': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '50'})
        }
    }

    complete_apps = ['avocado']



from django import forms
from django.db import models

from avocado.criteria import mixins
from avocado.concepts.models import Concept, ConceptField
from avocado.fields.models import Field
from avocado.criteria.viewset import library

__all__ = ('Criterion', 'CriterionField')

default = None
if library.default:
    default = library._get_class_name(library.default.__class__)

class Criterion(Concept, mixins.Mixin):
    fields = models.ManyToManyField(Field, through='CriterionField')
    viewset = models.CharField(max_length=100, choices=sorted(library.choices()),
        default=default)

    class Meta(Concept.Meta):
        verbose_name_plural = 'criteria'

    def _get_form(self):
        if not hasattr(self, '_form'):
            from avocado.criteria.cache import cache
            form_fields = {}
            fields = cache.get_fields(self.id)

            for f in fields:
                form_fields[f.field_name] = f.formfield()

            class CriterionForm(forms.Form):
                def __init__(self, *args, **kwargs):
                    super(CriterionForm, self).__init__(*args, **kwargs)
                    self.fields.update(form_fields)

            self._form = CriterionForm
        return self._form
    form = property(_get_form)

    def view_responses(self):
        resps = library.get(self.viewset, self)
        return resps


class CriterionField(ConceptField):
    concept = models.ForeignKey(Criterion, related_name='conceptfields')
    field = models.ForeignKey(Field, limit_choices_to={'is_public': True})
    required = models.BooleanField(default=True)

    class Meta(ConceptField.Meta):
        pass

    def text(self, operator, value):
        return ('%s %s' % (self.get_name(), operator.text(value))).strip()

    def get_name(self):
        return self.name or self.field.name



from django import forms
from django.db import models

from avocado.fields.models import Field

class FieldAdminForm(forms.ModelForm):

    def clean_app_name(self):
        app_name = self.cleaned_data['app_name']
        if models.get_app(app_name) is None:
            raise forms.ValidationError, 'The app "%s" could not be found' % app_name
        return app_name

    def clean(self):
        cleaned_data = super(FieldAdminForm, self).clean()

        app_name = cleaned_data['app_name']

        # test `model_name'
        model_name = cleaned_data['model_name']
        if self.instance._get_model(app_name, model_name) is None:
            del cleaned_data['model_name']
            msg = 'The model "%s" could not be found' % model_name
            self._errors['model_name'] = self.error_class([msg])

        # test `field_name'
        field_name = cleaned_data['field_name']
        if self.instance._get_field(field_name) is None:
            del cleaned_data['field_name']
            msg = 'The model "%s" does not have a field named "%s"' % \
                (model_name, field_name)
            self._errors['field_name'] = self.error_class([msg])

        # test `choices_handler'
        choices_handler = cleaned_data.get('choices_handler', None)
        if choices_handler and not self.instance._get_choices(choices_handler):
            del cleaned_data['choices_handler']
            msg = 'The choices could not be evaluated'
            self._errors['choices_handler'] = self.error_class([msg])

        return cleaned_data

    class Meta(object):
        model = Field



from django.db import models
from django.utils.importlib import import_module

from avocado.conf import settings

# attempt to import the user-defined mixin class if specified
if settings.FIELD_MIXIN_PATH is not None:
    path = settings.FIELD_MIXIN_PATH.split('.')
    mixin_name = path.pop(-1)
    mod = import_module('.'.join(path))
    Mixin = getattr(mod, mixin_name)
else:
    Mixin = models.Model



from functools import wraps

from avocado.utils.iter import ins

def _check_iter(func):
    @wraps(func)
    def decorator(*args, **kwargs):
        val = func(*args, **kwargs)
        if not ins(val):
            raise TypeError
        for e in val:
            if not ins(e) or len(e) != 2:
                raise TypeError
        return tuple(val)
    return decorator

@_check_iter
def eval_choices(val):
    # try evaling a straight sequence in the format:
    #   [(1,'foo'), (2,'bar'), ...]
    try:
        return eval(val)
    except (SyntaxError, NameError):
        return None

@_check_iter
def model_attr(model, attr):
    # attempts to check the `model' for an attribute `attr':
    #   when: attr = SHAPE_CHOICES
    #   test: model.SHAPE_CHOICES
    return getattr(model, attr, None)

@_check_iter
def module_attr(module, attr):
    return getattr(module, attr, None)


def evaluate(mf):
    ch = mf.choices_handler
    funcs = (
        (model_attr, (mf.model, ch)),
        (module_attr, (mf.module, ch)),
        (eval_choices, (ch,)),
    )

    for func, attrs in funcs:
        try:
            choices = func(*attrs)
            break
        except TypeError:
            pass
    else:
        choices = None
    return choices



from django.db import models
from django.utils.importlib import import_module

from avocado.conf import settings
from avocado.utils.mixins import create_mixin
from avocado.columns.format import library

# attempt to import the user-defined mixin class if specified
if settings.COLUMN_MIXIN_PATH is not None:
    path = settings.COLUMN_MIXIN_PATH.split('.')
    mixin_name = path.pop(-1)
    mod = import_module('.'.join(path))
    Mixin = getattr(mod, mixin_name)
else:
    Mixin = models.Model

# XXX ghetto...
default = None
if library.default:
    default = library._get_class_name(library.default.__class__)

fields = {}
for name in settings.FORMATTER_TYPES:
    fn = name + settings.FORMATTER_FIELD_SUFFIX
    fields[fn] = models.CharField('%s formatter' % name, max_length=100,
        choices=sorted(library.choices(name)), default=default)

Mixin = create_mixin('Mixin', __name__, bases=(Mixin,), fields=fields)



from copy import deepcopy
from django.template import defaultfilters as filters
from avocado.conf import settings
from avocado.exceptions import AlreadyRegisteredError
from avocado.concepts.library import Library
from avocado.utils.iter import ins

class FormatError(Exception):
    pass


class AbstractFormatter(object):
    """The AbstractFormatter defines the base behavior for formatting a set of
    objects.

    Each ``AbstractFormatter`` subclass will be registered with each output
    type if it has an applicable method. For example, if I have defined two
    output types 'json' and 'csv', then the following class will only be
    registered with the 'json' output type::

        class MyFormatter(AbstractFormatter):
            def json(self, *args, **args):
                # ...

    The requirement is that a method on the formatter subclass with the same
    name as the output type is defined. To override this behavior and apply
    a formatter to all output types, set ``apply_to_all`` to true.

    .. note::

        Note that when the formatters are registered, a single object is
        instantiated and registered with each output type. The objects should
        be stateless since the object can be shared.

    """
    apply_to_all = False
    is_choice = True

    def __call__(self, ftype, *args, **kwargs):
        return getattr(self, ftype)(*args, **kwargs)


class RemoveFormatter(AbstractFormatter):
    "Returns an empty list, removing the output."
    apply_to_all = True
    is_choice = False

    def __call__(self, ftype, *args, **kwargs):
        return []


class PassFormatter(AbstractFormatter):
    "Overrides default behavior. Simple retruns the args as-is."
    apply_to_all = True
    is_choice = False

    def __call__(self, ftype, *args, **kwargs):
        return args


class DefaultFormatter(AbstractFormatter):
    def html(self, *args, **kwargs):
        toks = []
        for x in args:
            if x is None: continue
            if type(x) is float:
                t = filters.floatformat(x)
            else:
                t = str(x)
            toks.append(t)
        return ' '.join(toks) or None

    def csv(self, *args, **kwargs):
        return args or None


class FormatterLibrary(Library):
    """The base class for defining a formatter library.

    This library dynamically determines the available formatter types via a
    user-defined setting `COLUMN_CONCEPT_FORMATTER_TYPES'. This provides the
    available to define any number of formatter types without needing to
    subclass or override any part of the library. An example as follows:

        FORMATTER_TYPES = {
            'json': {
                'error': '[data format error]',
            },
            'html' : {
                'error': '<span class="data-format-error">[data format error]</span>',
                'null': '<span class="lg ht">(no data)</span>',
            },
            'csv': {
                'error': '[data format error]',
                'null': '',
            }
        }

    Every formatter registered must be a sublcass of the AbstractFormatter
    class. Each formatter can support any number of formatter types by simply
    defining a method of the same name, e.g::

        class MyFormatter(AbstractFormatter):
            def json(*args):
                return ' '.join(map(lambda x: str(x).upper(), args))

    The above formatter class supports the "json" formatter type but not
    either of the "html" or "csv" types. For formatters that need to support
    more than one type, but handles the input the same, the following
    convention should be used:

        class MyFormatter(AbstractFormatter):
            def json(*args):
                return ' '.join(map(lambda x: str(x).upper(), args))
            html = json

    The "html" formatter merely points to the "json" method.
    """
    superclass = AbstractFormatter
    module_name = settings.FORMATTER_MODULE_NAME
    suffix = 'Formatter'

    def __init__(self, *args, **kwargs):
        super(FormatterLibrary, self).__init__(*args, **kwargs)
        self._cache = deepcopy(settings.FORMATTER_TYPES)
        self.format_types = self._cache.keys()

        for key in self._cache:
            self._cache[key]['formatters'] = {}

    def add_object(self, key, class_name, obj, errmsg=''):
        store = self.get_store()[key]['formatters']
        # if already registered under the same name, test if the classes are
        # the same -- replace if true, throw error if not
        if store.has_key(class_name):
            robj = store.get(class_name)
            if obj.__class__ is not robj.__class__:
                raise AlreadyRegisteredError, errmsg
        store.update({class_name: obj})

    def register_object(self, klass_name, obj):
        for key in self.format_types:
            if hasattr(obj, key) or obj.apply_to_all:
                self.add_object(key, klass_name, obj)

    def choices(self, key):
        "Returns a list of tuples that can be used as choices in a form."
        store = self.get_store()[key]['formatters']
        choices = []

        for name, obj in store.items():
            if obj.is_choice:
                choices.append((name, name))
        return choices

    def format_seq(self, seq, rules, ftype, formatters, error, null, **kwargs):
        index, toks = 0, []

        for fname, length in rules:
            part = seq[index:index+length]

            # if the formatter expects one argument, but it is None,
            # skip the formatting and default to None.
            if len(part) == 1 and part[0] is None:
                tok = None
            else:
                try:
                    obj = formatters[fname]
                    tok = obj(ftype, *part, **kwargs)
                except Exception:
                    tok = error

            if ins(tok):
                tok = list(tok)
                for i, t in enumerate(iter(tok)):
                    if t is None:
                        tok[i] = null
                toks.extend(tok)
            else:
                if tok is None:
                    tok = null
                toks.append(tok)

            index += length

        # all args have not been processed, therefore mixed formatting may have
        # occurred.
        if index != len(seq):
            raise FormatError, 'The rules "%s" is being applied to a ' \
                'sequence of %d items' % (rules, len(seq))

        return tuple(toks)

    def format(self, iterable, rules, ftype, **kwargs):
        """Take an iterable and formats the data given the rules.

        Definitions:

            `iterable' - an iterable of iterables

            `rules' - a list of pairs defining the formatter name and the
            number of items to format, e.g. [('MyFormatter', 3),
            ('AnotherFormatter', 1)]

            `ftype' - a string specifying the formatter type

                class MyFormatter(AbstractFormatter):
                    def json(*args):
                        return ' '.join(map(lambda x: str(x).upper(), args))

                iterable    = [(3, 'foo', 746, 'bar', 392)]
                rules       = (2, 'My')
                ftype       = 'json'

                The result will be:

                    [(3, 'FOO', '848', 'BAR', 392)]
        """
        store = self.get_store()[ftype]

        formatters = store['formatters']
        error = store.get('error', '')
        null = store.get('null', None)

        for seq in iter(iterable):
            yield self.format_seq(seq, rules, ftype, formatters, error, null, **kwargs)


library = FormatterLibrary()

library.register(RemoveFormatter)
library.register(PassFormatter)
library.register(DefaultFormatter)

library.default = DefaultFormatter()

library.autodiscover()



from django import forms
from django.db.models import Q

from avocado.exceptions import ValidationError
from avocado.conf import settings
from avocado.concepts.library import Library
from avocado.fields.operators import MODEL_FIELD_MAP
from avocado.modeltree import DEFAULT_MODELTREE_ALIAS
from avocado.utils.iter import ins

class OperatorNotPermitted(Exception):
    pass


class AbstractTranslator(object):
    "The base translator class that all translators must subclass."
    operators = None
    formfield = None

    formfield_overrides = {
        'IntegerField': forms.FloatField,
        'AutoField': forms.IntegerField
    }

    def __call__(self, field, operator=None, value=None, using=DEFAULT_MODELTREE_ALIAS, **context):
        return self.translate(field, operator, value, using=using, **context)

    def _clean_operator(self, field, operator, **kwargs):
        if self.operators:
            operators = self.operators
        else:
            operators = MODEL_FIELD_MAP.get(field.field.__class__.__name__)
        operators = dict([(x.uid, x) for x in operators])

        if not operators.has_key(operator):
            raise OperatorNotPermitted, 'operator "%s" cannot be used for this translator' % operator
        return operators[operator]

    def _clean_value(self, field, value, **kwargs):
        formfield = None

        # has highest precedence if defined on the class. this is usually NOT
        # necessary
        if self.formfield:
            formfield = self.formfield
        # special case for handling standalone ``None`` or ``bool`` values.
        # this occurs when a field is can be queried as a null value, i.e.
        # '-?isnull' : (True|False) or '-?exact' : None
        elif value is None or type(value) is bool:
            formfield = forms.NullBooleanField

        # create an instance of the formfield "to be" and determine if there is
        # a mapping listed for it. TODO make more elegant
        else:
            formfield = field.formfield

            if formfield() is None:
                name = field.field.__class__.__name__
            else:
                name = formfield().__class__.__name__

            if self.formfield_overrides.has_key(name):
                formfield = self.formfield_overrides[name]

        # TODO since None is considered an empty value by the django validators
        # ``required`` has to be set to False to not raise a ValidationError
        # saying the field is required. There may be a need to more explicitly
        # check to see if the value be passed is only None and not any of the
        # other empty values in ``django.core.validators.EMPTY_VALUES``
        ff = field.formfield(formfield=formfield, required=False, **kwargs)

        # special case for ``None`` values since all form fields seem to handle
        # the conversion differently. simply ignore the cleaning if ``None``,
        # this scenario occurs when a list of values are being queried and one
        # of them is to lookup NULL values
        if ins(value):
            new_value = []
            unique_values = set([])
            for x in value:
                if x in unique_values:
                    continue
                unique_values.add(x)
                if x is not None:
                    new_value.append(ff.clean(x))
                # Django assumes an empty string when given a ``NoneType``
                # for char-based form fields, this is to ensure ``NoneType``
                # are passed through unmodified
                else:
                    new_value.append(None)
            return new_value
        return ff.clean(value)

    def _get_not_null_pk(self, field, using):
        # XXX the below logic is required to get the expected results back
        # when querying for NULL values. since NULL can be a value and a
        # placeholder for non-existent values, then a condition to ensure
        # the row's primary key is also NOT NULL must be added. Django
        # assumes in all cases when querying on a NULL value all joins in
        # the chain up to that point must be promoted to LEFT OUTER JOINs,
        # which could be a reasonable assumption for some cases, but for
        # getting the existent rows of data back for our purposes, the
        # assumption is wrong.

        # if this field is already the primary key, then don't bother
        # adding the test, since it would be redundant
        if field.field.primary_key:
            return Q()

        from avocado.models import Field
        pk_name = field.model._meta.pk.name

        pk_field = Field(app_name=field.app_name,
            model_name=field.model_name, field_name=pk_name)

        key = pk_field.query_string('isnull', using=using)

        return Q(**{key: False})

    def _condition(self, field, operator, value, using):
        # assuming the operator and value validate, check for a NoneType value
        # if the operator is 'in'. This condition will be broken out into a
        # separate Q object
        if operator.operator == 'in' and None in value:
            value = value[:]
            value.remove(None)

            key = field.query_string('isnull', using=using)

            if operator.negated:
                condition = Q(**{key: False})
            else:
                condition = Q(**{key: True}) & self._get_not_null_pk(field, using)

            if value:
                key = field.query_string(operator.operator, using=using)
                condition = Q(**{key: value}) | condition
        else:
            if (operator.operator == 'isnull' or
                (operator.operator == 'exact' and value is None)):

                key = field.query_string('isnull', using=using)
                # This was is -isnull, thus we need to switch the value as well
                if operator.negated:
                    value = not value

                condition = Q(**{key: value})

                if value is True:
                    condition = condition & self._get_not_null_pk(field, using)
            else:
                key = field.query_string(operator.operator, using=using)
                condition = Q(**{key: value})

                condition = ~condition if operator.negated else condition

        return condition

    def validate(self, field, operator, value, **kwargs):
        clean_op = self._clean_operator(field, operator)
        clean_val = self._clean_value(field, value)
        if not clean_op.check(clean_val):
            raise ValidationError, '"%s" is not valid for the operator "%s"' % (clean_val, clean_op)
        return clean_op, clean_val

    def translate(self, field, roperator, rvalue, using, **context):
        """Returns two types of queryset modifiers including:
            - a Q object applied via the `filter()' method
            - a dict of annotations

        It should be noted that no checks are performed to prevent the same
        name being used for annotations.
        """
        operator, value = self.validate(field, roperator, rvalue, **context)
        condition = self._condition(field, operator, value, using)

        meta = {
            'condition': condition,
            'annotations': {},
            'cleaned_data': {
                'operator': operator,
                'value': value
            },
            'raw_data': {
                'operator': roperator,
                'value': rvalue,
            }
        }

        return meta


class DefaultTranslator(AbstractTranslator):
    "Provides the default behavior of creating a simple lookup."
    pass


class TranslatorLibrary(Library):
    superclass = AbstractTranslator
    module_name = settings.TRANSLATOR_MODULE_NAME
    suffix = 'Translator'
    default = DefaultTranslator()


library = TranslatorLibrary()
library.autodiscover()



from django.template import Context
from django.template.loader import get_template

from avocado.models import Column, Criterion, Field

FIELD_TEMPLATE_PATH = 'avocado/field-doc'
CONCEPT_TEMPLATE_PATH = 'avocado/concept-doc'

def generate_field_doc(pk, template_name=None, mimetype='txt'):
    "Generates the documentation for a single public field"
    try:
        field = Field.objects.get(pk=pk)
    except Field.DoesNotExist:
        return ''

    if template_name is None:
        template_name = '%s.%s' % (FIELD_TEMPLATE_PATH, mimetype)

    t = get_template(template_name)
    c = Context({'field': field})

    return t.render(c)

def generate_fields_doc(template_name=None, mimetype='txt'):
    "Generates the documentation for all fields"
    content = ''
    queryset = Field.objects.public()

    for f in queryset:
        content += generate_field_doc(f.pk, template_name, mimetype)
    return content

def generate_criterion_doc(pk, template_name=None, mimetype='txt'):
    "Generates the documentation for a single public criterion"
    try:
        criterion = Criterion.objects.get(pk=pk)
    except Criterion.DoesNotExist:
        return ''

    if template_name is None:
        template_name = '%s.%s' % (CONCEPT_TEMPLATE_PATH, mimetype)

    t = get_template(template_name)
    c = Context({
        'concept': criterion
    })

    return t.render(c)

def generate_criteria_doc(template_name=None, mimetype='txt'):
    content = ''
    queryset = Criterion.objects.public().order_by('order')

    for c in queryset:
        content += generate_criterion_doc(c.pk, template_name, mimetype)
    return content

def generate_column_doc(pk, template_name=None, mimetype='txt'):
    "Generates the documentation for a single public column"
    try:
        column = Column.objects.get(pk=pk)
    except Column.DoesNotExist:
        return ''

    if template_name is None:
        template_name = '%s.%s' % (CONCEPT_TEMPLATE_PATH, mimetype)

    t = get_template(template_name)
    c = Context({
        'concept': column
    })

    return t.render(c)

    pass

def generate_columns_doc(template_name=None, mimetype='txt'):
    content = ''
    queryset = Column.objects.public().order_by('order')

    for c in queryset:
        content += generate_column_doc(c.pk, template_name, mimetype)
    return content



import os
import sys

from django.conf import settings
from django.db.models import get_app
from django.db.models.signals import post_syncdb
from django.db import connections, transaction, DEFAULT_DB_ALIAS

def load_custom_sql(app, created_models, verbosity=2, **kwargs):
    app_dir = os.path.normpath(os.path.join(os.path.dirname(app.__file__), 'sql'))
    backend = settings.DATABASES[DEFAULT_DB_ALIAS]['ENGINE'].split('.')[-1]

    custom_files = (os.path.join(app_dir, "custom.%s.sql" % backend),
                    os.path.join(app_dir, "custom.sql"))

    for custom_file in custom_files:
        if os.path.exists(custom_file):
            print "Loading customized SQL for %s" % app.__name__
            fp = open(custom_file, 'U')
            cursor = connections['default'].cursor()
            try:
                cursor.execute(fp.read().decode(settings.FILE_CHARSET))
            except Exception:
                sys.stderr.write("Couldn't execute custom SQL for %s" % app.__name__)
                import traceback
                traceback.print_exc()
                transaction.rollback_unless_managed()
            else:
                transaction.commit_unless_managed()

post_syncdb.connect(load_custom_sql, sender=get_app('avocado'))



# encoding: utf-8
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models

class Migration(SchemaMigration):

    def forwards(self, orm):
        
        # Adding model 'Category'
        db.create_table('avocado_category', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=100)),
            ('icon', self.gf('django.db.models.fields.files.FileField')(max_length=100, blank=True)),
        ))
        db.send_create_signal('avocado', ['Category'])

        # Adding model 'Field'
        db.create_table(u'avocado_field', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('chart_title', self.gf('django.db.models.fields.CharField')(max_length=100, null=True, blank=True)),
            ('chart_yaxis', self.gf('django.db.models.fields.CharField')(max_length=100, null=True, blank=True)),
            ('chart_xaxis', self.gf('django.db.models.fields.CharField')(max_length=100, null=True, blank=True)),
            ('app_name', self.gf('django.db.models.fields.CharField')(max_length=100)),
            ('model_name', self.gf('django.db.models.fields.CharField')(max_length=100)),
            ('field_name', self.gf('django.db.models.fields.CharField')(max_length=100)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=100)),
            ('description', self.gf('django.db.models.fields.TextField')(null=True, blank=True)),
            ('keywords', self.gf('django.db.models.fields.CharField')(max_length=100, null=True, blank=True)),
            ('search_doc', self.gf('django.db.models.fields.TextField')(null=True, db_index=True)),
            ('is_public', self.gf('django.db.models.fields.BooleanField')(default=False)),
            ('group', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['auth.Group'], null=True, blank=True)),
            ('translator', self.gf('django.db.models.fields.CharField')(max_length=100, null=True, blank=True)),
            ('enable_choices', self.gf('django.db.models.fields.BooleanField')(default=False)),
            ('choices_handler', self.gf('django.db.models.fields.TextField')(null=True, blank=True)),
        ))
        db.send_create_signal(u'avocado', ['Field'])

        # Adding unique constraint on 'Field', fields ['app_name', 'model_name', 'field_name']
        db.create_unique(u'avocado_field', ['app_name', 'model_name', 'field_name'])

        # Adding model 'Column'
        db.create_table('avocado_column', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=100)),
            ('description', self.gf('django.db.models.fields.TextField')(null=True, blank=True)),
            ('keywords', self.gf('django.db.models.fields.CharField')(max_length=100, null=True, blank=True)),
            ('category', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['avocado.Category'], null=True, blank=True)),
            ('is_public', self.gf('django.db.models.fields.BooleanField')(default=False)),
            ('order', self.gf('django.db.models.fields.FloatField')(default=0)),
            ('search_doc', self.gf('django.db.models.fields.TextField')(null=True)),
            ('html_fmtr', self.gf('django.db.models.fields.CharField')(max_length=100)),
            ('csv_fmtr', self.gf('django.db.models.fields.CharField')(max_length=100)),
        ))
        db.send_create_signal('avocado', ['Column'])

        # Adding model 'ColumnField'
        db.create_table('avocado_columnfield', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=100, null=True, blank=True)),
            ('order', self.gf('django.db.models.fields.FloatField')(default=0)),
            ('concept', self.gf('django.db.models.fields.related.ForeignKey')(related_name='conceptfields', to=orm['avocado.Column'])),
            ('field', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['avocado.Field'])),
        ))
        db.send_create_signal('avocado', ['ColumnField'])

        # Adding model 'Criterion'
        db.create_table('avocado_criterion', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=100)),
            ('description', self.gf('django.db.models.fields.TextField')(null=True, blank=True)),
            ('keywords', self.gf('django.db.models.fields.CharField')(max_length=100, null=True, blank=True)),
            ('category', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['avocado.Category'], null=True, blank=True)),
            ('is_public', self.gf('django.db.models.fields.BooleanField')(default=False)),
            ('order', self.gf('django.db.models.fields.FloatField')(default=0)),
            ('search_doc', self.gf('django.db.models.fields.TextField')(null=True)),
            ('viewset', self.gf('django.db.models.fields.CharField')(max_length=100)),
        ))
        db.send_create_signal('avocado', ['Criterion'])

        # Adding model 'CriterionField'
        db.create_table('avocado_criterionfield', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=100, null=True, blank=True)),
            ('order', self.gf('django.db.models.fields.FloatField')(default=0)),
            ('concept', self.gf('django.db.models.fields.related.ForeignKey')(related_name='conceptfields', to=orm['avocado.Criterion'])),
            ('field', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['avocado.Field'])),
            ('required', self.gf('django.db.models.fields.BooleanField')(default=True)),
        ))
        db.send_create_signal('avocado', ['CriterionField'])

        # Adding model 'Scope'
        db.create_table('avocado_scope', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('user', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['auth.User'], null=True, blank=True)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=100, null=True, blank=True)),
            ('description', self.gf('django.db.models.fields.TextField')(null=True, blank=True)),
            ('keywords', self.gf('django.db.models.fields.CharField')(max_length=100, null=True, blank=True)),
            ('store', self.gf('avocado.store.fields.PickledField')(null=True, editable=False)),
            ('definition', self.gf('django.db.models.fields.TextField')(null=True)),
            ('timestamp', self.gf('django.db.models.fields.DateTimeField')(default=datetime.datetime(2011, 2, 25, 13, 14, 9, 564798))),
            ('cnt', self.gf('django.db.models.fields.PositiveIntegerField')()),
        ))
        db.send_create_signal('avocado', ['Scope'])

        # Adding model 'Perspective'
        db.create_table('avocado_perspective', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('user', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['auth.User'], null=True, blank=True)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=100, null=True, blank=True)),
            ('description', self.gf('django.db.models.fields.TextField')(null=True, blank=True)),
            ('keywords', self.gf('django.db.models.fields.CharField')(max_length=100, null=True, blank=True)),
            ('store', self.gf('avocado.store.fields.PickledField')(null=True, editable=False)),
            ('definition', self.gf('django.db.models.fields.TextField')(null=True)),
            ('timestamp', self.gf('django.db.models.fields.DateTimeField')(default=datetime.datetime(2011, 2, 25, 13, 14, 9, 564798))),
        ))
        db.send_create_signal('avocado', ['Perspective'])

        # Adding model 'Report'
        db.create_table('avocado_report', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('user', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['auth.User'], null=True, blank=True)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=100, null=True, blank=True)),
            ('description', self.gf('django.db.models.fields.TextField')(null=True, blank=True)),
            ('keywords', self.gf('django.db.models.fields.CharField')(max_length=100, null=True, blank=True)),
            ('scope', self.gf('django.db.models.fields.related.OneToOneField')(to=orm['avocado.Scope'], unique=True)),
            ('perspective', self.gf('django.db.models.fields.related.OneToOneField')(to=orm['avocado.Perspective'], unique=True)),
        ))
        db.send_create_signal('avocado', ['Report'])


    def backwards(self, orm):
        
        # Removing unique constraint on 'Field', fields ['app_name', 'model_name', 'field_name']
        db.delete_unique(u'avocado_field', ['app_name', 'model_name', 'field_name'])

        # Deleting model 'Category'
        db.delete_table('avocado_category')

        # Deleting model 'Field'
        db.delete_table(u'avocado_field')

        # Deleting model 'Column'
        db.delete_table('avocado_column')

        # Deleting model 'ColumnField'
        db.delete_table('avocado_columnfield')

        # Deleting model 'Criterion'
        db.delete_table('avocado_criterion')

        # Deleting model 'CriterionField'
        db.delete_table('avocado_criterionfield')

        # Deleting model 'Scope'
        db.delete_table('avocado_scope')

        # Deleting model 'Perspective'
        db.delete_table('avocado_perspective')

        # Deleting model 'Report'
        db.delete_table('avocado_report')


    models = {
        'auth.group': {
            'Meta': {'object_name': 'Group'},
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '80'}),
            'permissions': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Permission']", 'symmetrical': 'False', 'blank': 'True'})
        },
        'auth.permission': {
            'Meta': {'ordering': "('content_type__app_label', 'content_type__model', 'codename')", 'unique_together': "(('content_type', 'codename'),)", 'object_name': 'Permission'},
            'codename': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['contenttypes.ContentType']"}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '50'})
        },
        'auth.user': {
            'Meta': {'object_name': 'User'},
            'date_joined': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'email': ('django.db.models.fields.EmailField', [], {'max_length': '75', 'blank': 'True'}),
            'first_name': ('django.db.models.fields.CharField', [], {'max_length': '30', 'blank': 'True'}),
            'groups': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Group']", 'symmetrical': 'False', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_active': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'is_staff': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'is_superuser': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'last_login': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'last_name': ('django.db.models.fields.CharField', [], {'max_length': '30', 'blank': 'True'}),
            'password': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'user_permissions': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Permission']", 'symmetrical': 'False', 'blank': 'True'}),
            'username': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '30'})
        },
        'avocado.category': {
            'Meta': {'ordering': "('name',)", 'object_name': 'Category'},
            'icon': ('django.db.models.fields.files.FileField', [], {'max_length': '100', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
        },
        'avocado.column': {
            'Meta': {'ordering': "('name',)", 'object_name': 'Column'},
            'category': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['avocado.Category']", 'null': 'True', 'blank': 'True'}),
            'csv_fmtr': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'description': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'fields': ('django.db.models.fields.related.ManyToManyField', [], {'to': u"orm['avocado.Field']", 'through': "orm['avocado.ColumnField']", 'symmetrical': 'False'}),
            'html_fmtr': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_public': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'keywords': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'order': ('django.db.models.fields.FloatField', [], {'default': '0'}),
            'search_doc': ('django.db.models.fields.TextField', [], {'null': 'True'})
        },
        'avocado.columnfield': {
            'Meta': {'ordering': "('order',)", 'object_name': 'ColumnField'},
            'concept': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'conceptfields'", 'to': "orm['avocado.Column']"}),
            'field': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['avocado.Field']"}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'order': ('django.db.models.fields.FloatField', [], {'default': '0'})
        },
        'avocado.criterion': {
            'Meta': {'ordering': "('name',)", 'object_name': 'Criterion'},
            'category': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['avocado.Category']", 'null': 'True', 'blank': 'True'}),
            'description': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'fields': ('django.db.models.fields.related.ManyToManyField', [], {'to': u"orm['avocado.Field']", 'through': "orm['avocado.CriterionField']", 'symmetrical': 'False'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_public': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'keywords': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'order': ('django.db.models.fields.FloatField', [], {'default': '0'}),
            'search_doc': ('django.db.models.fields.TextField', [], {'null': 'True'}),
            'viewset': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        'avocado.criterionfield': {
            'Meta': {'ordering': "('order',)", 'object_name': 'CriterionField'},
            'concept': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'conceptfields'", 'to': "orm['avocado.Criterion']"}),
            'field': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['avocado.Field']"}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'order': ('django.db.models.fields.FloatField', [], {'default': '0'}),
            'required': ('django.db.models.fields.BooleanField', [], {'default': 'True'})
        },
        u'avocado.field': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_name', 'model_name', 'field_name'),)", 'object_name': 'Field'},
            'app_name': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'chart_title': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'chart_xaxis': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'chart_yaxis': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'choices_handler': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'description': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'enable_choices': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'field_name': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'group': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['auth.Group']", 'null': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_public': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'keywords': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'model_name': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'search_doc': ('django.db.models.fields.TextField', [], {'null': 'True', 'db_index': 'True'}),
            'translator': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'})
        },
        'avocado.perspective': {
            'Meta': {'object_name': 'Perspective'},
            'definition': ('django.db.models.fields.TextField', [], {'null': 'True'}),
            'description': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'keywords': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'store': ('avocado.store.fields.PickledField', [], {'null': 'True', 'editable': 'False'}),
            'timestamp': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime(2011, 2, 25, 13, 14, 9, 564798)'}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['auth.User']", 'null': 'True', 'blank': 'True'})
        },
        'avocado.report': {
            'Meta': {'object_name': 'Report'},
            'description': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'keywords': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'perspective': ('django.db.models.fields.related.OneToOneField', [], {'to': "orm['avocado.Perspective']", 'unique': 'True'}),
            'scope': ('django.db.models.fields.related.OneToOneField', [], {'to': "orm['avocado.Scope']", 'unique': 'True'}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['auth.User']", 'null': 'True', 'blank': 'True'})
        },
        'avocado.scope': {
            'Meta': {'object_name': 'Scope'},
            'cnt': ('django.db.models.fields.PositiveIntegerField', [], {}),
            'definition': ('django.db.models.fields.TextField', [], {'null': 'True'}),
            'description': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'keywords': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'store': ('avocado.store.fields.PickledField', [], {'null': 'True', 'editable': 'False'}),
            'timestamp': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime(2011, 2, 25, 13, 14, 9, 564798)'}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['auth.User']", 'null': 'True', 'blank': 'True'})
        },
        'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        }
    }

    complete_apps = ['avocado']



from django.contrib import admin

from avocado.concepts.models import Category

__all__ = ('ConceptAdmin', 'EditorsConceptAdmin')

class ConceptAdmin(admin.ModelAdmin):
    list_display = ('name', 'category', 'is_public', 'order', 'field_relations')
    list_filter = ('is_public', 'category')
    list_editable = ('category', 'is_public', 'order')
    search_fields = ('name', 'description', 'keywords')
    ordering = ('name',)
    list_per_page = 25
    save_as = True


class EditorsConceptAdmin(admin.ModelAdmin):
    list_display = ('name', 'category', 'is_public')
    list_filter = ('is_public', 'category')
    search_fields = ('name', 'description', 'keywords')
    ordering = ('name',)
    list_per_page = 25

    readonly_fields = ('is_public',)

    fieldsets = (
        (None, {
            'fields': ('name', 'description', 'keywords', 'category',
                'is_public')
        }),
    )


class CategoryAdmin(admin.ModelAdmin):
    list_display = ('name', 'parent', 'order')
    list_editable = ('parent', 'order')


admin.site.register(Category, CategoryAdmin)



def ins(value):
    """Simple test to distinguish sequences that are not strings.

    >>> ins(None)
    False
    >>> ins('')
    False
    >>> ins(u'')
    False
    >>> ins(r'')
    False
    >>> ins(True)
    False

    >>> ins([])
    True
    >>> ins(())
    True
    >>> ins({})
    True
    >>> ins(set([]))
    True
    """
    if not isinstance(value, basestring):
        try:
            iter(value)
            return True
        except TypeError:
            pass
    return False



import cPickle as pickle
from hashlib import md5
from datetime import datetime
from functools import partial

from django.db import models, router
from django.db.models.sql import RawQuery
from django.core.paginator import EmptyPage, InvalidPage
from django.contrib.auth.models import User
from django.core.cache import cache as dcache

from forkit.models import ForkableModel
from forkit import signals

from avocado.conf import settings
from avocado.models import Field
from avocado.store.fields import JSONField
from avocado.modeltree import DEFAULT_MODELTREE_ALIAS, trees
from avocado.fields import logictree
from avocado.columns.cache import cache as column_cache
from avocado.columns import utils, format
from avocado.utils.paginator import BufferedPaginator
from avocado.store import receivers

__all__ = ('Scope', 'Perspective', 'Report', 'ObjectSet',
    'ObjectSetJoinThrough')

PAGE = 1
PAGINATE_BY = 10
CACHE_CHUNK_SIZE = 500
DEFAULT_COLUMNS = getattr(settings, 'COLUMNS', ())
DEFAULT_ORDERING = getattr(settings, 'COLUMN_ORDERING', ())

class Descriptor(ForkableModel):
    user = models.ForeignKey(User, null=True)
    name = models.CharField(max_length=100, null=True)
    description = models.TextField(null=True)
    keywords = models.CharField(max_length=100, null=True)
    created = models.DateTimeField(default=datetime.now)
    modified = models.DateTimeField(default=datetime.now)
    # explicitly denotes an instance for use in a session
    session = models.BooleanField(default=False)

    class Meta(object):
        abstract = True
        app_label = 'avocado'

    def save(self, *args, **kwargs):
        # since we want to mimic the reference object, we don't want to
        # update the modified date on save, otherwise they would not be
        # in a consistent state. this condition will never be true for
        # object's without a reference.
        if not self.session or self.has_changed():
            self.modified = datetime.now()
        super(Descriptor, self).save(*args, **kwargs)

    def __unicode__(self):
        if self.session:
            return u'{0} (session)'.format(self.name or self.pk)
        return u'{0}'.format(self.name or self.pk)

    def get_reference_pk(self):
        if self.reference:
            return self.reference.pk

    def references(self, pk):
        "Compares the reference primary key to the one passed."
        if self.reference:
            return self.reference.pk == int(pk)

    def deference(self, delete=False):
        if self.reference and delete:
            self.reference.delete()
        self.__class__().reset(self)
        self.reference = None
        self.save()

    def diff(self, instance=None, **kwargs):
        "Override diff to default to ``reference`` if no instance is sepcified."
        if not instance:
            if not self.reference:
                return None
            instance = self.reference
        return super(Descriptor, self).diff(instance, **kwargs)

    def push(self):
        "Pushes changes from this object to the reference, if one exists."
        self.reset(self.reference)

    def has_changed(self):
        return bool(self.diff())

    def model_name(self, using=DEFAULT_MODELTREE_ALIAS):
        return trees[using].root_model._meta.verbose_name.format()

    def model_name_plural(self, using=DEFAULT_MODELTREE_ALIAS):
        return trees[using].root_model._meta.verbose_name_plural.format()


class Context(Descriptor):
    """A generic interface for storing an arbitrary context around the data
    model. The object defining the context must be serializable.
    """
    store = JSONField(null=True)
    timestamp = models.DateTimeField(editable=False, auto_now=True, default=datetime.now)

    class Meta(object):
        abstract = True
        app_label = 'avocado'

    def _get_obj(self, obj=None):
        if obj is None:
            return self.store or {}
        return obj

    def _get_contents(self, obj):
        """A ``Context`` is driven by the abstraction layer of the ``Field``,
        ``Criterion`` and ``Column`` classes. Each ``obj`` will be empty or
        contain data (like primary keys) referring to objects of the former
        mentioned classes.

        Returns a list of ``Field`` primary keys.
        """
        pass

    def _parse_contents(self, obj, *args, **kwargs):
        """Encapsulates any processing that must be performed on ``obj`` and
        returns a function that takes a queryset and returns a queryset.
        """
        pass

    def cache_is_valid(self, timestamp=None):
        if timestamp and timestamp > self.timestamp:
            return True
        return False

    def is_valid(self, obj):
        """Takes an object and determines if the data structure is valid for
        this particular context.
        """
        if isinstance(obj, dict):
            return True
        return False

    def read(self):
        return self._get_obj()

    def write(self, obj=None, *args, **kwargs):
        obj = self._get_obj(obj)
        self.store = obj
        self.timestamp = datetime.now()

    def has_permission(self, obj=None, user=None):
        obj = self._get_obj(obj)

        field_ids = set([int(i) for i in self._get_contents(obj)])
        # if not requesting to see anything, early exit
        if not field_ids:
            return True

        fields = Field.objects.public(user)

        # filter down to requested fields
        ids = set(fields.values('id').filter(id__in=field_ids).values_list('id', flat=True))

        if len(ids) != len(field_ids) or not all([i in field_ids for i in ids]):
            return False

        return True

    def get_queryset(self, obj=None, queryset=None, using=DEFAULT_MODELTREE_ALIAS, *args, **kwargs):
        obj = self._get_obj(obj)

        if queryset is None:
            queryset = trees[using].get_queryset()

        func = self._parse_contents(obj, using=using, *args, **kwargs)
        queryset = func(queryset, *args, **kwargs)
        return queryset


class Scope(Context):
    "Stores information needed to provide scope to data."

    count = models.IntegerField(null=True, editable=False, db_column='cnt')
    # used for book keeping. if a reference exists, this implies this instance
    # has represents another context.
    reference = models.ForeignKey('self', null=True)


    def _get_contents(self, obj):
        return logictree.transform(obj).get_field_ids()

    def _parse_contents(self, obj, *args, **kwargs):
        return logictree.transform(obj).apply

    def save(self, *args, **kwargs):
        self.count = self.get_queryset().distinct().count()
        super(Scope, self).save(*args, **kwargs)

    @property
    def conditions(self):
        return logictree.transform(self._get_obj()).text


class Perspective(Context):
    # used for book keeping. if a reference exists, this implies this instance
    # has represents another context.
    reference = models.ForeignKey('self', null=True)

    def _get_obj(self, obj=None):
        obj = obj or {}
        if self.store is not None:
            copy = self.store.copy()
        else:
            copy = {}

        copy.update(obj)

        # supply default values
        if not copy.has_key('columns'):
            copy['columns'] = list(DEFAULT_COLUMNS)
        if not copy.has_key('ordering'):
            copy['ordering'] = list(DEFAULT_ORDERING)

        copy['columns'] = [int(x) for x in copy['columns']]
        copy['ordering'] = [(int(x), y) for x, y in copy['ordering']]

        # ordering of a column cannot exist when the column is not present
        for i, (x, y) in enumerate(iter(copy['ordering'])):
            if x not in copy['columns']:
                copy['ordering'].pop(i)

        return copy

    def _get_contents(self, obj):
        ids = obj['columns'] + [x for x,y in obj['ordering']]

        # saves a query
        if not ids:
            return ids

        # get all field ids associated with requested columns
        return Field.objects.filter(column__id__in=set(ids)).values_list('id',
            flat=True)

    def _parse_contents(self, obj, *args, **kwargs):
        def func(queryset, columns=[], ordering=[], *args, **kwargs):
            queryset = utils.add_columns(queryset, columns, *args, **kwargs)
            queryset = utils.add_ordering(queryset, ordering, *args, **kwargs)
            return queryset

        return partial(func, columns=obj['columns'], ordering=obj['ordering'])

    def header(self):
        store = self.read()
        header = []

        for x in store['columns']:
            c = column_cache.get(x)
            if c is None:
                continue
            o = {'id': x, 'name': c.name, 'direction': ''}
            for y, z in store['ordering']:
                if x == y:
                    o['direction'] = z
                    break
            header.append(o)

        return header

    def get_columns_as_fields(self):
        store = self.read()
        header = []

        for x in store['columns']:
            c = column_cache.get(x)
            cfields = c.conceptfields.select_related('field').order_by('order')
            if len(cfields) == 1:
                header.append(c.name)
            else:
                header.extend([x.name or x.field.name for x in cfields])
        return header

    def format(self, iterable, format_type, **kwargs):
        store = self.read()
        rules = utils.column_format_rules(store['columns'], format_type)
        return format.library.format(iterable, rules, format_type, **kwargs)


class Report(Descriptor):
    "Represents a combination ``scope`` and ``perspective``."
    REPORT_CACHE_KEY = 'reportcache'

    scope = models.OneToOneField(Scope)
    perspective = models.OneToOneField(Perspective)
    # used for book keeping. if a reference exists, this implies this instance
    # has represents another context.
    reference = models.ForeignKey('self', null=True)
    count = models.IntegerField(null=True, editable=False, db_column='cnt')

    def _center_cache_offset(self, count, offset, buf_size=CACHE_CHUNK_SIZE):
        """The ``offset`` will be relative to the next requested row. To ensure
        a true 'sliding window' of data, the offset must be adjusted to be::

            offset - (buf_size / 2)

        The edge cases will be relative to the min (0) and max number of rows
        that exist.
        """
        mid = int(buf_size / 2.0)

        # lower bound
        if (offset - mid) < 0:
            offset = 0
        # upper bound
        elif (offset + mid) > count:
            offset = count - buf_size
        # in the middle
        else:
            offset = offset - mid

        return offset

    def _set_queryset_offset_limit(self, queryset, offset, limit):
        lower = offset
        upper = offset + limit
        return queryset[lower:upper]

    def _execute_raw_query(self, queryset):
        """Take a ``QuerySet`` object and executes it. No customization or
        processing of the query should take place here.
        """
        using = router.db_for_read(queryset.model)
        sql, params = queryset.query.get_compiler(using).as_sql()
        raw = RawQuery(sql, using, params)
        raw._execute_query()
        return raw.cursor.fetchall()

    def deference(self, delete=False):
        self.scope.deference()
        self.perspective.deference()
        if self.reference and delete:
            self.reference.delete()
        self.__class__().reset(self)
        self.reference = None
        self.save()

    def paginator_and_page(self, cache, buf_size=CACHE_CHUNK_SIZE):
        paginator = BufferedPaginator(count=cache['count'], offset=cache['offset'],
            buf_size=buf_size, per_page=cache['per_page'])

        try:
            page = paginator.page(cache['page_num'])
        except (EmptyPage, InvalidPage):
            page = paginator.page(paginator.num_pages)

        return paginator, page

    def get_datakey(self, request):
        return md5(request.session._session_key + 'data').hexdigest()

    def cache_is_valid(self, timestamp=None):
        if self.scope.cache_is_valid(timestamp) and \
            self.perspective.cache_is_valid(timestamp):
            return True
        return False

    # in it's current implementation, this will try to get the requested
    # page from cache, or re-execute the query and store off the new cache
    def get_page_from_cache(self, cache, buf_size=CACHE_CHUNK_SIZE):
        paginator, page = self.paginator_and_page(cache, buf_size)

        # now we can fetch the data
        if page.in_cache():
            data = dcache.get(cache['datakey'])
            if data is not None:
                return page.get_list(pickle.loads(data))

    def refresh_cache(self, cache, queryset, adjust_offset=True, buf_size=CACHE_CHUNK_SIZE):
        """Does not utilize existing cache if it exists. This is an implied
        cache invalidation mechanism.
        """
        paginator, page = self.paginator_and_page(cache, buf_size)

        queryset = self._set_queryset_offset_limit(queryset, cache['offset'], buf_size)

        # since the page is not in cache new data must be requested, therefore
        # the offset should be re-centered relative to the page offset
        if adjust_offset:
            cache['offset'] = self._center_cache_offset(cache['count'], page.offset(), buf_size)

        data = self._execute_raw_query(queryset)
        dcache.set(cache['datakey'], pickle.dumps(data))

        paginator.offset = cache['offset']
        paginator.object_list = data

        try:
            page = paginator.page(cache['page_num'])
        except (EmptyPage, InvalidPage):
            page = paginator.page(paginator.num_pages)

        assert page.in_cache()

        return page.get_list()

    def update_cache(self, cache, queryset, buf_size=CACHE_CHUNK_SIZE):
        """Tries to use cache if it exists, this implies that the cache is still
        valid and a page that is not in cache has been requested.
        """
        paginator, page = self.paginator_and_page(cache, buf_size)

        # since the page is not in cache new data must be requested, therefore
        # the offset should be re-centered relative to the page offset
        cache['offset'] = self._center_cache_offset(cache['count'], page.offset(), buf_size)

        # determine any overlap between what we have with ``cached_rows`` and
        # what the ``page`` requires.
        has_overlap, start_term, end_term = paginator.get_overlap(cache['offset'])

        # we can run a partial query and use some of the existing rows for our
        # updated cache
        if has_overlap is False:
            queryset = self._set_queryset_offset_limit(queryset, *start_term)
            data = self._execute_raw_query(queryset)
        else:
            rdata = dcache.get(cache['datakey'])
            if rdata is None:
                return self.refresh_cache(cache, queryset, adjust_offset=False,
                    buf_size=buf_size)

            data = pickle.loads(rdata)
            # check to see if there is partial data to be prepended
            if start_term[0] is not None:
                tmp = self._set_queryset_offset_limit(queryset, *start_term)
                partial_data = self._execute_raw_query(tmp)
                data = partial_data + data[:-start_term[1]]

            # check to see if there is partial data to be appended
            if end_term[0] is not None:
                tmp = self._set_queryset_offset_limit(queryset, *end_term)
                partial_data = self._execute_raw_query(tmp)
                data = data[end_term[1]:] + partial_data

        dcache.set(cache['datakey'], pickle.dumps(data))

        paginator.offset = cache['offset']
        paginator.object_list = data

        page = paginator.page(cache['page_num'])
        assert page.in_cache()

        return page.get_list()

    def _get_count(self, queryset):
        tmp = queryset.all()
        tmp.query.clear_ordering(True)
        return tmp.count()

    def get_queryset(self, timestamp=None, using=DEFAULT_MODELTREE_ALIAS, **context):
        """Returns a ``QuerySet`` object that is generated from the ``scope``
        and ``perspective`` objects bound to this report. This should not be
        used directly when requesting data since it does not utlize the cache
        layer.
        """
        unique = count = None
        queryset = trees[using].get_queryset()
        queryset = queryset.values(queryset.model._meta.pk.name).distinct()

        # first argument is ``None`` since we want to use the session objects
        queryset = self.scope.get_queryset(None, queryset, using=using, **context)
        unique = self._get_count(queryset)

        queryset = self.perspective.get_queryset(None, queryset, using=using)
        count = self._get_count(queryset)

        if self.count != count:
            self.count = count
            self.save()

        return queryset, unique, count

    def has_permission(self, user):
        # ensure the requesting user has permission to view the contents of
        # both the ``scope`` and ``perspective`` objects
        # TODO add per-user caching for report objects
        if self.scope.has_permission(user=user) and self.perspective.has_permission(user=user):
            return True
        return False

    def has_changed(self):
        return self.scope.has_changed() or self.perspective.has_changed()

class ObjectSet(Descriptor):
    """
    Provides a means of saving off a set of objects.

    `criteria' is persisted so the original can be rebuilt. `removed_ids'
    is persisted to know which objects have been excluded explicitly from the
    set. This could be useful when testing for if there are new objects
    available when additional data has been loaded, while still excluding
    the removed objects.

    `ObjectSet' must be subclassed to add the many-to-many relationship
    to the "object" of interest.
    """
    scope = models.OneToOneField(Scope, editable=False)
    count = models.PositiveIntegerField(null=True, editable=False, db_column='cnt')

    class Meta(object):
        abstract = True

    def save(self, *args, **kwargs):
        if not self.created:
            self.created = datetime.now()
        self.modified = datetime.now()
        super(ObjectSet, self).save(*args, **kwargs)


class ObjectSetJoinThrough(models.Model):
    """
    Adds additional information about the objects that have been ``added`` and
    ``removed`` from the original set.

    For instance, additional objects that are added which do not match the
    conditions currently associated with the ``ObjectSet`` should be flagged
    as ``added``. If in the future they match the conditions, the flag can be
    removed.

    Any objects that are removed from the set should be marked as ``removed``
    even if they were added at one time. This is too keep track of the objects
    that have been explicitly removed from the set.
    """
    removed = models.BooleanField(default=False)
    added = models.BooleanField(default=False)

    class Meta(object):
        abstract = True


signals.pre_diff.connect(receivers.descriptor_pre_diff, sender=Scope)
signals.pre_reset.connect(receivers.descriptor_pre_reset, sender=Scope)
signals.pre_fork.connect(receivers.descriptor_pre_fork, sender=Scope)
signals.post_fork.connect(receivers.descriptor_post_fork, sender=Scope)
signals.pre_commit.connect(receivers.descriptor_pre_commit, sender=Scope)
signals.post_commit.connect(receivers.descriptor_post_commit, sender=Scope)

signals.pre_diff.connect(receivers.descriptor_pre_diff, sender=Perspective)
signals.pre_reset.connect(receivers.descriptor_pre_reset, sender=Perspective)
signals.pre_fork.connect(receivers.descriptor_pre_fork, sender=Perspective)
signals.post_fork.connect(receivers.descriptor_post_fork, sender=Perspective)
signals.pre_commit.connect(receivers.descriptor_pre_commit, sender=Perspective)
signals.post_commit.connect(receivers.descriptor_post_commit, sender=Perspective)

signals.pre_diff.connect(receivers.report_pre_diff, sender=Report)
signals.pre_reset.connect(receivers.report_pre_reset, sender=Report)
signals.pre_fork.connect(receivers.report_pre_fork, sender=Report)
signals.post_fork.connect(receivers.descriptor_post_fork, sender=Report)
signals.pre_commit.connect(receivers.descriptor_pre_commit, sender=Report)
signals.post_commit.connect(receivers.descriptor_post_commit, sender=Report)



from avocado.columns.models import Column
from avocado.columns.cache import cache
from avocado.modeltree import DEFAULT_MODELTREE_ALIAS

def column_format_rules(columns, format_type, pk_pass=True):
    if pk_pass:
        rules = [('Pass', 1)]
    else:
        rules = []

    for column in columns:
        if not isinstance(column, Column):
            column = cache.get(column)
        rules.append(column.rule(format_type))
    return rules

def add_columns(queryset, columns, using=DEFAULT_MODELTREE_ALIAS):
    """Takes a ``queryset`` and ensures the proper table join have been
    setup to display the table columns.
    """
    # add queryset model's pk field
    aliases = [(queryset.model._meta.db_table,
        queryset.model._meta.pk.column)]

    for column in columns:
        if not isinstance(column, Column):
            column = cache.get(column)
        queryset, _aliases = column.add_fields_to_queryset(queryset, using)
        aliases.extend(_aliases)

    queryset.query.select = aliases
    return queryset

def add_ordering(queryset, order_rules, using=DEFAULT_MODELTREE_ALIAS):
    """Applies column ordering to a queryset. Resolves a Column's
    fields and generates the `order_by' paths.
    """
    queryset.query.clear_ordering(True)

    if not order_rules:
        return queryset

    orders = []

    for column, direction in order_rules:
        if not isinstance(column, Column):
            column = cache.get(column)
        _orders = column.get_ordering_for_queryset(direction, using=using)
        orders.extend(_orders)

    return queryset.order_by(*orders)



def descriptor_pre_diff(sender, reference, instance, config, **kwargs):
    config['exclude'] = ('pk', 'reference', 'session', 'modified', 'created', 'timestamp')

def descriptor_pre_reset(sender, reference, instance, config, **kwargs):
    config['exclude'] = ('pk', 'reference', 'session')

def descriptor_pre_fork(sender, reference, instance, config, **kwargs):
    config['exclude'] = ('pk', 'reference', 'session', 'user')
    config['deep'] = False

def descriptor_post_fork(sender, reference, instance, **kwargs):
    instance.user = reference.user

def descriptor_pre_commit(sender, reference, instance, **kwargs):
    # if the instance being committed is the session object, update
    # the reference
    if instance.session:
        instance.reference = reference

def descriptor_post_commit(sender, reference, instance, **kwargs):
    # this looks odd, but in the case of the persistent session object,
    # `reference' must keep a reference to the instance it is mimicking
    if reference.session:
        reference.reference = instance
        reference.commit()


def report_pre_diff(sender, reference, instance, config, **kwargs):
    if not config.get('fields', None):
        config['fields'] = ('name', 'description', 'keywords', 'scope', 'perspective')
    config['deep'] = True

def report_pre_reset(sender, reference, instance, config, **kwargs):
    config['exclude'] = ('pk', 'reference', 'session')
    config['deep'] = True

def report_pre_fork(sender, reference, instance, config, **kwargs):
    # reference and report_set are the same things, due to the self reference..
    config['exclude'] = ('pk', 'reference', 'session', 'user', 'report_set')
    config['deep'] = True



"""
The LogicTree provides a means of parsing a raw JSON-compatible structure. The
grammar is as follows:

    NODE    ::= F_NODE | L_NODE

    F_NODE  ::= '{'
                    'id'        ::= INT
                    'operator'  ::= STRING
                    'value'     ::= STRING | NUMBER | BOOL | None
                '}'

    L_NODE  ::= '{'
                    'type'  ::= 'AND' | 'OR'
                    'children'  ::= '[' NODE{2,} ']'
                '}'

Two examples as follows:

    # single F_NODE
    {
        'id': 1,
        'operator': 'exact',
        'value': 30
    }

    # nested L_NODE structure
    {
        'type': 'OR',
        'children': [{
            'id': 1,
            'operator': 'exact',
            'value': 30
        }, {
            'id': 1,
            'operator': 'exact',
            'value': 45
        }]
    }
"""
from avocado.modeltree import DEFAULT_MODELTREE_ALIAS
from avocado.models import Field, Criterion, CriterionField

AND = 'AND'
OR = 'OR'

class Node(object):
    condition = None
    annotations = None

    def get_field_ids(self):
        return []

    def apply(self, queryset, *args, **kwargs):
        if self.annotations:
            queryset = queryset.values('pk').annotate(**self.annotations)
        if self.condition:
            queryset = queryset.filter(self.condition)
        return queryset

    @property
    def text(self, *args, **kwargs):
        pass


class Condition(Node):
    "Contains information for a single query condition."
    def __init__(self, context=None, using=DEFAULT_MODELTREE_ALIAS, **kwargs):
        self.context = context
        self.using = using
        self.id = kwargs['id']
        self.operator = kwargs['operator']
        self.value = kwargs['value']
        self.concept_id = kwargs['concept_id']

    @property
    def _meta(self):
        if not hasattr(self, '__meta'):
            self.__meta = self.field.translate(self.operator, self.value,
                using=self.using, **self.context)
        return self.__meta

    @property
    def criterion(self):
        if not hasattr(self, '_criterion'):
            self._criterion = Criterion.objects.get(id=self.concept_id)
        return self._criterion

    @property
    def criterionfield(self):
        if not hasattr(self, '_criterionfield'):
            self._criterionfield = CriterionField.objects.get(concept__id=self.concept_id,
                field__id=self.id)
        return self._criterionfield

    @property
    def field(self):
        if not hasattr(self, '_field'):
            self._field = Field.objects.get(id=self.id)
        return self._field

    @property
    def condition(self):
        return self._meta['condition']

    @property
    def annotations(self):
        return self._meta['annotations']

    @property
    def text(self, flatten=True):
        operator = self._meta['cleaned_data']['operator']
        # the original value is used here to prevent representing a different
        # value from what the client had submitted. this text has no impact
        # on the stored 'cleaned' data structure
        value = self._meta['raw_data']['value']
        return {'conditions': [self.criterionfield.text(operator, value)]}

    def get_field_ids(self):
        return [self.id]


class LogicalOperator(Node):
    "Provides a logical relationship between it's children."
    def __init__(self, type, using=DEFAULT_MODELTREE_ALIAS):
        self.using = using
        self.type = (type.upper() == AND) and AND or OR
        self.children = []

    def _combine(self, q1, q2):
        if self.type.upper() == OR:
            return q1 | q2
        return q1 & q2

    @property
    def condition(self):
        if not hasattr(self, '_condition'):
            condition = None
            for node in self.children:
                if condition:
                    condition = self._combine(node.condition, condition)
                else:
                    condition = node.condition
            self._condition = condition
        return self._condition

    @property
    def annotations(self):
        if not hasattr(self, '_annotations'):
            self._annotations = {}
            for node in self.children:
                self._annotations.update(node.annotations)
        return self._annotations

    @property
    def text(self, flatten=True):
        if not hasattr(self, '_text'):
            text = {'type': self.type.lower(), 'conditions': []}
            for node in self.children:
                t = node.text
                # flatten if nested conditions are also AND's
                if not t.has_key('type') or t['type'] == self.type.lower():
                    text['conditions'].extend(t['conditions'])
                else:
                    text['conditions'].append(t['conditions'])
            self._text = text
        return self._text

    def get_field_ids(self):
        ids = []
        for node in self.children:
            ids.extend(node.get_field_ids())
        return ids


def transform(rnode, pnode=None, using=DEFAULT_MODELTREE_ALIAS, **context):
    "Takes the raw data structure and converts it into the node tree."
    if not rnode:
        return Node()

    if rnode.has_key('children'):
        # ensure the logic makes sense
        if len(rnode['children']) < 2:
            raise TypeError, 'a logical operator must apply to 2 or more ' \
                'conditions'
        node = LogicalOperator(rnode['type'], using=using)
        for child in rnode['children']:
            transform(child, node, using=using, **context)
    else:
        node = Condition(context, using=using, **rnode)
    # top level node returns, i.e. no parent node
    if pnode is None:
        return node
    pnode.children.append(node)



from django import forms
from django.db import transaction
from avocado.store.models import Report
from .context import ContextForm, SessionContextForm
from .scope import ScopeForm
from .perspective import PerspectiveForm

class ReportForm(ContextForm):
    scope = forms.Field(required=False)
    perspective = forms.Field(required=False)

    def clean_scope(self):
        if self.instance.pk:
            data = self.cleaned_data.get('scope')
            if data:
                form = ScopeForm(data, instance=self.instance.scope)
                if form.is_valid():
                    form.save(commit=False)
            return self.instance.scope

    def clean_perspective(self):
        if self.instance.pk:
            data = self.cleaned_data.get('perspective')
            if data:
                form = PerspectiveForm(data, instance=self.instance.perspective)
                if form.is_valid():
                    form.save(commit=False)
            return self.instance.perspective

    @transaction.commit_on_success
    def save(self, commit=True):
        instance = super(ReportForm, self).save(commit=False)
        if commit:
            instance.scope.save()
            instance.perspective.save()
            instance.save()
        return instance

    class Meta(ContextForm.Meta):
        model = Report
        fields = ('name', 'description', 'keywords', 'scope', 'perspective')


class SessionReportForm(SessionContextForm):
    class Meta(object):
        model = Report
        fields = ('name', 'description', 'keywords')



from django.db import transaction
from django import forms

class ContextForm(forms.ModelForm):
    name = forms.CharField(required=False)
    description = forms.CharField(required=False)
    keywords = forms.CharField(required=False)
    store = forms.Field(required=False)

    class Meta(object):
        fields = ('name', 'description', 'keywords', 'store')


class SessionContextForm(ContextForm):
    @transaction.commit_on_success
    def save(self, commit=True):
        # shortcut for deferencing and clearing the session instance
        if not self.data and self.instance:
            # deference already saves the instance
            self.instance.deference()
            return self.instance

        # apply changes to the session object
        instance = super(SessionContextForm, self).save(commit=False)

        # if no reference, but a name has been specified create a fork
        if not instance.reference and instance.name:
            return instance.fork()
        elif instance.diff(fields=['name', 'description', 'keywords']):
            # if a reference does exist and both the name and other stuff
            # has changed create a fork (not to overwrite an existing report)
            if instance.has_changed():
                return instance.fork()
            # if it's just metadata that has changed, update the reference
            # in-place
            else:
                instance.push()

        instance.save()
        return instance

    class Meta(object):
        fields = ('name', 'description', 'keywords', 'store')



from avocado.concepts.cache import ConceptCache
from avocado.columns.models import Column

class ColumnCache(ConceptCache):
    def get(self, concept_id, queryset=None, ret_val=None):
        if queryset is None:
            queryset = Column.objects.all()
        return super(ColumnCache, self).get(concept_id=concept_id,
            queryset=queryset, ret_val=ret_val)

    def get_fields(self, concept_id, queryset=None, ret_val=None):
        if queryset is None:
            queryset = Column.objects.all()
        return super(ColumnCache, self).get_fields(concept_id=concept_id,
            queryset=queryset, ret_val=ret_val)


cache = ColumnCache(Column.__name__.lower())



from avocado.store.models import Scope
from .context import ContextForm, SessionContextForm

class ScopeForm(ContextForm):
    class Meta(ContextForm.Meta):
        model = Scope


class SessionScopeForm(SessionContextForm):
    class Meta(SessionContextForm.Meta):
        model = Scope



from django.core.exceptions import ObjectDoesNotExist
from django.core.cache import cache as _cache

from avocado.fields.models import Field

class FieldCache(object):
    def __init__(self, class_name):
        self.id_key = ':'.join(['avocado', class_name, '%s'])
        self.label_key = ':'.join(['avocado', class_name, '%s', '%s', '%s'])
        self.field_id_key = ':'.join(['avocado', class_name, '%s', 'fields'])
        self.field_label_key = ':'.join(['avocado', class_name, '%s', '%s', '%s', 'fields'])
    
    def get(self, field_id=None, app_name=None, model_name=None, field_name=None, queryset=None, ret_val=None):
        "Simple interface for getting (and setting) a field from global cache."
        if not field_id and not (app_name and model_name and field_name):
            raise RuntimeError, 'not enough lookup params defined'

        if queryset is None:
            queryset = Field.objects.all()

        if field_id:
            key = self.id_key % field_id
            kwargs = {'id': field_id}
        else:
            key = self.label_key % (app_name, model_name, field_name)
            kwargs = {'app_name': app_name, 'model_name': model_name,
                'field_name': field_name}

        field = _cache.get(key, None)
        if field is None:
            try:
                field = queryset.get(**kwargs)
            except ObjectDoesNotExist:
                return ret_val
            # create other key and set a reference to the object for both keys
            if field_id:
                skey = self.label_key % (field.app_name, field.model_name,
                    field.field_name)
            else:
                skey = self.id_key % field.id

            _cache.set(key, field)
            _cache.set(skey, field)

        return field

    def get_many(self, ids_or_labels, queryset=None):
        "Returns a generator of field objects."
        for args in iter(ids_or_labels):
            if type(args) is int or str(args).isdigit():
                yield self.get(args, queryset=queryset)
            else:
                yield self.get(None, *args, queryset=queryset)


cache = FieldCache(Field.__name__.lower())



from django.db import models

def create_mixin(name, module, bases=None, fields=None, options=None):
    bases = bases or (models.Model,)

    class Meta(object):
        abstract = True

    # Update Meta with any options that were provided
    if options is not None:
        for key, value in options.iteritems():
            setattr(Meta, key, value)

    # Set up a dictionary to simulate declarations within a class
    attrs = {'__module__': module, 'Meta': Meta}

    # Add in any fields that were provided
    if fields:
        attrs.update(fields)

    # Create the class, which automatically triggers ModelBase processing
    model = type(name, bases, attrs)

    return model



CHARS = '63VhSRo0yHvg1upE4sDGXqjkx7Fd8KfmIPBJlU52brtaQciZM9NTLOnCzewYWA'

def encode(num, chars=CHARS):
    if num == 0:
        return chars[0]
    l = []
    base = len(chars)
    while num:
        rem = num % base
        num = num // base
        l.append(chars[rem])
    l.reverse()
    return ''.join(l)

def decode(string, chars=CHARS):
    base = len(chars)
    strlen = len(string)
    num = 0

    idx = 0
    for char in string:
        power = (strlen - (idx + 1))
        num += chars.index(char) * (base ** power)
        idx += 1

    return num




import inspect

from avocado.conf import settings
from avocado.concepts.library import Library

class AbstractViewSet(object):
    "The AbstractViewSet class."
    order = ()
    js = ''
    css = ''

    def __call__(self, concept, *args, **kwargs):
        cfields = list(concept.conceptfields.order_by('order'))
        return self._get_responses(concept, cfields, *args, **kwargs)

    def _get_responses(self, concept, cfields, *args, **kwargs):
        views = []
        resps = {}

        for name in dir(self):
            if name.startswith('_'):
                continue

            method = getattr(self, name)
            if inspect.ismethod(method):
                resp = method(concept, cfields, *args, **kwargs)
                if not resp.has_key('tabname'):
                    resp['tabname'] = name.replace('_', ' ').title()
                resps[name] = resp

        if self.order:
            for x in self.order:
                views.append(resps.pop(x))

        # add rest of un-ordered views
        views.extend(resps.values())

        resp = {
            'id': concept.id,
            'name': concept.name,
            'js': self.js or None,
            'css': self.css or None,
            'views': views,
        }

        return resp

class ViewSetLibrary(Library):
    superclass = AbstractViewSet
    module_name = settings.VIEWSET_MODULE_NAME
    suffix = 'ViewSet'

    def get(self, name, concept):
        viewset = super(ViewSetLibrary, self).get(name)
        return viewset(concept)


library = ViewSetLibrary()
library.autodiscover()



import re
import math

from django import forms
from django.db import models
from django.db.models import Count
from django.contrib.auth.models import Group
from django.contrib.sites.models import Site
from django.db.models.fields import FieldDoesNotExist

from avocado.modeltree import DEFAULT_MODELTREE_ALIAS, trees
from avocado.fields.translate import library
from avocado.fields.managers import FieldManager
from avocado.fields import mixins

__all__ = ('Field',)

COERCED_DATATYPES = (
    (re.compile(r'^integer|float|decimal|big|positive|small|auto'), 'number'),
    (re.compile(r'^char|text|file|ipaddress|slug'), 'string'),
)

class Field(mixins.Mixin):
    """The `Field' class stores off meta data about a "field of
    interest" located on another model. This, in a sense, provides a way to
    specify the fields that can be utilized by the query engine.

    There are three cases in which a "field of interest" should be stored:

        defintion/vocabulary - at the very least, storing off a field provides
        the ability specify a `description' and `keywords' (or aliases) associated
        with the field.

        queryability - the field can be associated with one or more
        `Criterions'. at minimum, this provides the ability to query by
        this field. in a more complex scenario, the field can act as a dependent
        or dependency on other fields.

        reporting - the field can be associated with one or more `Columns'
        which allows for generating reports (results) in-browser or exporting to
        another format.
    """
    app_name = models.CharField(max_length=100)
    model_name = models.CharField(max_length=100)
    field_name = models.CharField(max_length=100)

    name = models.CharField(max_length=100)
    description = models.TextField(null=True, blank=True)
    keywords = models.CharField(max_length=100, null=True, blank=True)

    # the ``search_doc`` refers to the aggregated text that can be fulltext
    # searched. if the database backend supports it, this will also be used
    # by the fulltext indexing to construct the search index
    search_doc = models.TextField(null=True, db_index=True, editable=False)

    is_public = models.BooleanField(default=False)
    group = models.ForeignKey(Group, null=True, blank=True)
    sites = models.ManyToManyField(Site, blank=True)

    translator = models.CharField(max_length=100, choices=library.choices(),
        blank=True, null=True)

    # specify a constrained list of choices for this field, applies for
    # valiation and for display in the UI
    enable_choices = models.BooleanField(default=False)
    choices_handler = models.TextField(null=True, blank=True, help_text="""
        Allowed callbacks include specifying:
            1. a constant name on the model
            2. a constant name on the model's module
            3. a string that can be evaluated
    """)

    objects = FieldManager()

    class Meta(object):
        app_label = 'avocado'
        unique_together = ('app_name', 'model_name', 'field_name')
        ordering = ('name',)

    def __unicode__(self):
        if self.name:
            name = '%s [%s]' % (self.name, self.model_name)
        else:
            name = '.'.join([self.app_name, self.model_name, self.field_name])
        return u'%s' % name

    def natural_key(self):
        return (self.app_name, self.model_name, self.field_name)

    def _get_module(self):
        "Helper for determining ``choices``, if ``choices_handler`` is defined."
        if not hasattr(self, '_module'):
            self._module = __import__(self.model.__module__)
        return self._module
    module = property(_get_module)

    def _get_model(self, app_name=None, model_name=None):
        "Returns the model class this field is associated with."
        if not hasattr(self, '_model') or (app_name and model_name):
            app_name = app_name or self.app_name
            model_name = model_name or self.model_name
            self._model = models.get_model(app_name, model_name)
        return self._model
    model = property(_get_model)

    def _get_field(self, field_name=None):
        "Returns the field object from model."
        if not hasattr(self, '_field') or field_name:
            field_name = field_name or self.field_name
            try:
                self._field = self.model._meta.get_field_by_name(field_name)[0]
            except FieldDoesNotExist:
                self._field = None
        return self._field
    field = property(_get_field)

    def _get_datatype(self):
        if not hasattr(self, '_datatype'):
            internal = self.field.get_internal_type().lower()
            for p, t in COERCED_DATATYPES:
                if p.match(internal):
                    self._datatype = t
                    break
            else:
                if internal.endswith('field'):
                    internal = internal[:-5]
                self._datatype = internal
        return self._datatype
    datatype = property(_get_datatype)

    def _get_choices(self, choices_handler=None):
        """Returns a distinct set of choices for this field.

        If a ``choices_handler`` is not defined, a distinct list of values are
        retrieved from the database.

        If ``choices_handler`` is defined, the handler is passed through a series
        of evaluators to try and convert it into a native object.
        """
        if not hasattr(self, '_choices') or choices_handler:
            choices = None

            # override boolean type fields
            if self.datatype in ('boolean', 'nullboolean'):
                choices = [(True, 'Yes'), (False, 'No')]
                if self.datatype == 'nullboolean':
                    choices.append((None, 'No Data'))
                choices = tuple(choices)

            elif self.enable_choices:
                choices_handler = choices_handler or self.choices_handler

                # use introspection
                if not choices_handler:
                    name = self.field_name
                    choices = list(self.model.objects.values_list(name,
                        flat=True).order_by(name).distinct())
                    choices = zip(choices, map(lambda x: x is None and 'No Data' or str(x), choices))

                # attempt to evaluate custom handler
                else:
                    from avocado.fields import evaluators
                    choices = evaluators.evaluate(self)

            self._choices = choices
        return self._choices
    choices = property(_get_choices)

    def _get_values(self):
        "Returns a list of the values."
        if self.choices is not None:
            return tuple(map(lambda x: x[0], self.choices))
        return ()
    values = property(_get_values)

    def distribution(self, exclude=[], min_count=None, max_points=100,
        order_by='field', smooth=0.01, annotate_by='pk', **filters):

        """Builds a GROUP BY queryset for use as a value distribution.
        Data is binned according to a bin width specified by the 
        Freedman-Diaconis Rule of h = 2 * IQR / n^(1/3) where
        h = bin width, IQR = Interquartile range, n = number of observations.
        Citation:
        Freedman, David; Diaconis, Persi (December 1981). "On the
        histogram as a density estimator: L2 theory" Probability
        Theory and Related Fields 57 (4): 453-476. ISSN 0178-8951.

        ``exclude`` - a list of values to be excluded from the distribution. it
        may be desired to exclude NULL values or the empty string.

        .. note::

            the default behavior of the ``exclude`` argument is to do a SQL
            equivalent of NOT IN (...). if ``None`` is included, it will have
            a custom behavior of IS NOT NULL and will be removed from the IN
            clause. default is to include all values

        ``max_points`` - the maximum number of points to be include in the
        distribution. the min and max values are always included, then a random
        sample is taken from the distribution. default is 100

        ``order_by`` - specify an ordering for the distribution. the choices are
        'count', 'field', or None. default is 'count'

        ``filters`` - a dict of filters to be applied to the queryset before
        the count annotation.
        """
        name = str(self.field_name)

        # get base queryset
        dist = self.model.objects.values(name)

        # exclude certain values (e.g. None, '')
        if exclude:
            exclude = set(exclude)
            kwargs = {}

            # special case for null values
            if None in exclude:
                kwargs['%s__isnull' % name] = True
                exclude.remove(None)

            kwargs['%s__in' % name] = exclude
            dist = dist.exclude(**kwargs)

        # apply filters before annotation is made
        if filters:
            dist = dist.filter(**filters)

        n = dist.count()

        if n >= 2 and self.datatype == 'number' and smooth >= 0:
            # evaluate
            dist = dist.values_list(name, flat=True)
            # raw ordered data
            dist = dist.order_by(name)

            # Bins are calculated using the Freedman-Diaconis' method.
            # F-D Method is:
            # h = 2 * (IQR / n^(1/3)) where
            # h = bin width
            # IQR is the interquartile region. Iqr = the differenc in 
            # the third and first quartiles. 
            # n is the number of data points
            # This can be changed if a better
            # method is found or could be a
            # parameter choice.
            # Citation:
            # Freedman, David; Diaconis, Persi (December 1981). "On the
            # histogram as a density estimator: L2 theory" Probability
            # Theory and Related Fields 57 (4): 453-476. ISSN 0178-8951.
            first_quartile = 0.25
            third_quartile = 0.75
            q1 = dist[int(math.ceil(n * first_quartile))]
            q3 = dist[int(math.floor(n * third_quartile))]
            iqr = q3 - q1
            h = 2 * (float(iqr) * pow(n, -(1.0 / 3.0)))

            dist = dist.annotate(count=Count(name)).values_list(name, 'count')
            minimum_pt = dist[0]
            maximum_pt = dist.reverse()[0]
            bin_data = []
            current_bin = float(minimum_pt[0]) + h
            bin_height = 0
            for data_pt in dist.iterator():
                if data_pt in [minimum_pt, maximum_pt]:
                    continue
                pt = float(data_pt[0])
                # If data point is less than the current bin
                # add to the bin height
                if pt <= current_bin:
                    bin_height += data_pt[1]
                if pt > current_bin:
                    x = current_bin
                    y = bin_height
                    prev = (0, 0)
                    if bin_data:
                        prev = bin_data.pop()

                    # compare current bin to previous 
                    # if prev bin is too small, the current
                    # bin takes in previous.
                    # Previous bin takes in current bin, 
                    # if current bin is too small
                    if y > 0:
                        if (y*smooth) > prev[1]:
                            fact = prev[1] / y
                            bin_x = x - (h / 2) - fact
                            bin_y = y + prev[1]
                            xy = (bin_x, bin_y)
                        elif prev[1]*smooth > y:
                            fact = y / prev[1]
                            bin_x = prev[0] + fact
                            bin_y = y + prev[1]
                            xy = (bin_x, bin_y)
                        else:
                            bin_data.append(prev)
                            bin_x = x - (h / 2)
                            xy = (bin_x, y)
                        bin_data.append(xy)
                    bin_height = 0

                    # increment to next bin until data_pt
                    # is within a bin. Add to height and 
                    # move to next data_pt
                    if h == 0:
                        return [(0, 0)]
                    while pt > current_bin:
                        current_bin += h
                    bin_height += data_pt[1]
            # Add back the min and max points and return the
            # list of X, Y coordinates.
            bin_data.insert(0, (float(minimum_pt[0]), minimum_pt[1]))
            bin_data.append((float(maximum_pt[0]), maximum_pt[1]))
            return bin_data


        # This only applies to catagorical data

        # apply annotation
        if annotate_by == 'pk':
            annotate_by = self.model._meta.pk.name
        dist = dist.annotate(count=Count(annotate_by))

        # evaluate
        dist = dist.values_list(name, 'count')

        # apply ordering
        if order_by == 'count':
            dist = dist.order_by('count')
        elif order_by == 'field':
            dist = dist.order_by(name)

        return tuple(dist)

    def query_string(self, operator=None, using=DEFAULT_MODELTREE_ALIAS):
        "Returns a django lookup string relative to the ``modeltree`` object."
        modeltree = trees[using]
        nodes = modeltree.path_to(self.model)
        return modeltree.query_string(nodes, self.field_name, operator)

    def order_string(self, direction='asc', using=DEFAULT_MODELTREE_ALIAS):
        qs = self.query_string(using=using)
        if direction.lower() == 'desc':
            return '-' + qs
        return qs

    def translate(self, operator=None, value=None, using=DEFAULT_MODELTREE_ALIAS, **context):
        trans = library.get(self.translator)
        return trans(self, operator, value, using, **context)

    def query_by_value(self, operator, value, using=DEFAULT_MODELTREE_ALIAS):
        modeltree = trees[using]
        meta = self.translate(operator, value, using)
        queryset = modeltree.root_model.objects.all()

        if meta['annotations']:
            queryset = queryset.annotate(**meta['annotations'])
        if meta['condition']:
            queryset = queryset.filter(meta['condition'])
        return queryset

    def formfield(self, formfield=None, widget=None, **kwargs):
        "Returns the default `formfield' instance for the `field' type."
        if formfield is None:
            formfield = self.field.formfield

        if 'label' in kwargs:
            label = kwargs.pop('label')
        else:
            label = self.name.title()

        if not widget and self.enable_choices:
            widget = forms.SelectMultiple(choices=self.choices)

        return formfield(label=label, widget=widget, **kwargs)




"""
Simple set of classes that roughly map one-to-one to the operations that can
performed in the django ORM.

Each class must provide a `clean' method that checks a given `value' is of
the right length and in some cases, type.
"""
import re

from avocado.utils.iter import ins

__all__ = ('exact', 'iexact', 'contains', 'inlist', 'lt', 'gt', 'lte', 'gte',
    'between', 'null', 'notbetween', 'notexact', 'notiexact', 'doesnotcontain',
    'notinlist', 'notnull', 'MODEL_FIELD_MAP')

FIELD_LOOKUPS = re.compile(r'(i?(exact|contains|regex|startswidth|endswith)|'\
    '(gt|lt)e?|in|range|year|month|day|week_day|isnull|search)')

TEXT_MAX_LIST_SIZE = 3

class Operator(object):
    short_name = ''
    verbose_name = ''
    operator = ''
    negated = False

    def __str__(self):
        if self.negated:
            return '%s (~%s)' % (self.short_name, self.operator)
        return '%s (%s)' % (self.short_name, self.operator)

    def __unicode__(self):
        return u'%s' % str(self)

    def __repr__(self):
        return str(self.__class__)

    def _get_uid(self):
        if self.negated:
            return '-%s' % self.operator
        return '%s' % self.operator
    uid = property(_get_uid)

    def stringify(self, value):
        s = '<b>%s</b>'
        if value is None:
            if self.negated:
                value = 'has any value'
            else:
                value = 'has no value'
        elif type(value) is bool:
            value = value and 'Yes' or 'No'
        else:
            value = str(value)

        return s % value

    def check(self, value):
        pass

    def text(self, value):
        pass


class PrimitiveOperator(Operator):
    def check(self, value):
        if not ins(value):
            return True
        return False

    def text(self, value):
        value = self.stringify(value)
        return '%s %s' % (self.verbose_name, value)


class SequenceOperator(Operator):
    join_operator = 'or'

    def check(self, value):
        if ins(value):
            return True
        return False

    def text(self, value):
        value = map(self.stringify, value)

        last = value[-1]
        length = len(value)-1

        if length == 0:
            if self.negated:
                return '%s %s' % (NotExact.verbose_name, last)
            return '%s %s' % (Exact.verbose_name, last)

        if length > TEXT_MAX_LIST_SIZE:
            head = value[:TEXT_MAX_LIST_SIZE]
        else:
            head = value[:-1]
        text = self.verbose_name + ' ' + ', '.join(head)

        # Add the leftover item count for the tail of the list
        tail = length - TEXT_MAX_LIST_SIZE
        if tail > 0:
            text += ' ... (%s more)' % tail

        text += (' %s ' % self.join_operator) + last

        return text


class Exact(PrimitiveOperator):
    "Only used with boolean fields. Use `iexact' otherwise."
    short_name = '='
    verbose_name = 'is equal to'
    operator = 'exact'
exact = Exact()


class iExact(PrimitiveOperator):
    short_name = '='
    verbose_name = 'is equal to'
    operator = 'iexact'
iexact = iExact()


class Contains(PrimitiveOperator):
    short_name = 'contains'
    verbose_name = 'contains the text'
    operator = 'icontains'
contains = Contains()


class LessThan(PrimitiveOperator):
    short_name = '<'
    verbose_name = 'is less than'
    operator = 'lt'
lt = LessThan()


class GreaterThan(PrimitiveOperator):
    short_name = '>'
    verbose_name = 'is greater than'
    operator = 'gt'
gt = GreaterThan()


class LessThanOrEqual(PrimitiveOperator):
    short_name = '<='
    verbose_name = 'is less than or equal to'
    operator = 'lte'
lte = LessThanOrEqual()


class GreaterThanOrEqual(PrimitiveOperator):
    short_name = '>='
    verbose_name = 'is greater than or equal to'
    operator = 'gte'
gte = GreaterThanOrEqual()


class Null(PrimitiveOperator):
    short_name = 'is null'
    verbose_name = 'is null'
    operator = 'isnull'

    def text(self, value):
        "Do not return operator"
        return self.stringify(None)
null = Null()


class InList(SequenceOperator):
    short_name = 'in list'
    verbose_name = 'is either'
    operator = 'in'
inlist = InList()


class Between(SequenceOperator):
    join_operator = 'and'
    short_name = 'between'
    verbose_name = 'is between'
    operator = 'range'

    def check(self, value):
        if ins(value) and len(value) == 2:
            return True
        return False
between = Between()


class NotBetween(Between):
    short_name = 'not between'
    verbose_name = 'is not between'
    negated = True
notbetween = NotBetween()


class NotExact(Exact):
    short_name = '!='
    verbose_name = 'is not equal to'
    negated = True
notexact = NotExact()


class NotiExact(iExact):
    short_name = '!='
    verbose_name = 'is not equal to'
    negated = True
notiexact = NotiExact()


class DoesNotContain(Contains):
    short_name = 'does not contain'
    verbose_name = 'does not contain'
    negated = True
doesnotcontain = DoesNotContain()


class NotInList(InList):
    short_name = 'not in list'
    verbose_name = 'is not'
    negated = True
notinlist = NotInList()


class NotNull(Null):
    short_name = 'not null'
    verbose_name = 'is not null'
    negated = True
notnull = NotNull()

NULL_OPERATORS = (null, notnull)
GENERAL_OPERATORS = (inlist, notinlist, exact, notexact)

CHAR_OPERATORS = GENERAL_OPERATORS + NULL_OPERATORS + (iexact, notiexact,
    contains, doesnotcontain)

NUMERIC_OPERATORS = GENERAL_OPERATORS + NULL_OPERATORS + (lt, lte, gt, gte,
    between, notbetween)

MODEL_FIELD_MAP = {
    'AutoField': GENERAL_OPERATORS,
    'CharField': CHAR_OPERATORS,
    'TextField': CHAR_OPERATORS,
    'IntegerField': NUMERIC_OPERATORS,
    'FloatField': NUMERIC_OPERATORS,
    'DecimalField': NUMERIC_OPERATORS,
    'DateField': NUMERIC_OPERATORS,
    'DateTimeField': NUMERIC_OPERATORS,
    'TimeField': NUMERIC_OPERATORS,
    'BooleanField': GENERAL_OPERATORS,
    'NullBooleanField': NULL_OPERATORS + GENERAL_OPERATORS
}



from django import forms
from django.db import models
from django.core.urlresolvers import reverse
from django.contrib import admin

from avocado.concepts.admin import ConceptAdmin
from avocado.criteria.models import CriterionField, Criterion

__all__ = ('CriterionAdmin',)

class CriterionFieldInline(admin.TabularInline):
    model = CriterionField


class CriterionAdmin(ConceptAdmin):
    inlines = (CriterionFieldInline,)
    formfield_overrides = {
        models.TextField: {'widget': forms.Textarea(attrs={'cols': 30, 'rows': 3}),
            'required': False}
    }

    def field_relations(self, obj):
        queryset = obj.fields.order_by('criterionfield__order').only('id', 'name')
        urlize = lambda x: '<a href="%s">%s</a>' % (
            reverse('admin:avocado_field_change', args=(x.id,)),
            x.name
        )
        return '<br>'.join(map(urlize, queryset)) or None
    field_relations.short_description = 'Field Relations'
    field_relations.allow_tags = True


admin.site.register(Criterion, CriterionAdmin)



from django.contrib import admin
from django.db import transaction
from django.core.urlresolvers import reverse

from avocado.concepts.admin import ConceptAdmin, EditorsConceptAdmin
from avocado.models import Field, Column, Criterion
from avocado.fields.forms import FieldAdminForm

__all__ = ('FieldAdmin', 'EditorsFieldAdmin')


class FieldAdmin(ConceptAdmin):
    form = FieldAdminForm

    list_display = ('name', 'is_public', 'show_orphan_reason', 'model_name',
        'enable_choices', 'criterion_relations', 'column_relations')
    list_filter = ('is_public', 'model_name', 'sites')
    list_editable = ('is_public', 'enable_choices')

    actions = ('create_criterion', 'create_column')

    fieldsets = (
        (None, {
            'fields': ('name', 'description', 'keywords', 'is_public'),
        }),

        ('Query Representation', {
            'classes': ('collapse',),
            'fields': ('translator', 'enable_choices')
        }),

        ('Permissions', {
            'classes': ('collapse',),
            'fields': ('group', 'sites')
        }),

        ('Metadata', {
            'description': ('These fields should not be altered in-place. '
                'If where this field is located has changed, change these '
                ' values and create a new field by using "Save as new".'),
            'classes': ('collapse',),
            'fields': ('app_name', 'model_name', 'field_name'),
        }),
    )

    def show_orphan_reason(self, obj):
        if obj.model is None:
            return 'Unknown Model'
        if obj.field is None:
            return 'Unknown Field'
        return 'OK'
    show_orphan_reason.short_description = 'Orphan Status'

    def criterion_relations(self, obj):
        queryset = obj.criterion_set.only('id', 'name')
        urlize = lambda x: '<a href="%s">%s</a>' % (
            reverse('admin:avocado_criterion_change', args=(x.id,)),
            x.name
        )
        return '<br>'.join(map(urlize, queryset)) or None
    criterion_relations.short_description = 'Criterion Relations'
    criterion_relations.allow_tags = True

    def column_relations(self, obj):
        queryset = obj.column_set.only('id', 'name')
        urlize = lambda x: '<a href="%s">%s</a>' % (
            reverse('admin:avocado_column_change', args=(x.id,)),
            x.name
        )
        return '<br>'.join(map(urlize, queryset)) or None
    column_relations.short_description = 'Column Relations'
    column_relations.allow_tags = True

    @transaction.commit_on_success
    def _create_concept(self, model, request, queryset):
        for f in queryset:
            concept = model(name=f.name, is_public=False)
            concept.save()

            conceptfield = concept.conceptfields.model(field=f,
                concept=concept)
            conceptfield.save()

    def create_criterion(self, request, queryset):
        return self._create_concept(Criterion, request, queryset)
    create_criterion.short_description = 'Create criterion from field'

    def create_column(self, request, queryset):
        return self._create_concept(Column, request, queryset)
    create_column.short_description = 'Create column from field'


class EditorsFieldAdmin(EditorsConceptAdmin):
    list_display = ('name', 'is_public')
    list_filter = ('is_public',)

    fieldsets = (
        (None, {
            'fields': ('name', 'description', 'keywords', 'is_public')
        }),
    )


admin.site.register(Field, FieldAdmin)



import sys
import locale
import codecs
from optparse import make_option

from django.core.management.base import BaseCommand

from avocado import docs

class Command(BaseCommand):
    help = "Generates a report of all the metadata that exists."

    option_list = BaseCommand.option_list + (
        make_option('-t', '--type', dest='mimetype', default='txt',
            help='Specifies which template type to use e.g. text or html'),
    )

    def handle(self, *args, **options):
        output = docs.generate_fields_doc(mimetype=options['mimetype'])
        sys.stdout = codecs.getwriter(locale.getpreferredencoding())(sys.stdout)
        sys.stdout.write(output)
        sys.exit(0)



"""
The Column class takes an optional setting, 'COLUMN_CONCEPT_MIXIN' which
defines an abstract Model class that can provide references to additional
fields.

The most typical scenario for using this setting is to specify 'formatter'
fields for the Column class to use.
"""

from django.db import models

from avocado.conf import settings
from avocado.modeltree import DEFAULT_MODELTREE_ALIAS, trees
from avocado.concepts.models import Concept, ConceptField
from avocado.fields.models import Field
from avocado.columns import mixins

__all__ = ('Column', 'ColumnField')

class Column(Concept, mixins.Mixin):
    "An interface to specify the necessary fields for a column."
    fields = models.ManyToManyField(Field, through='ColumnField')

    class Meta(Concept.Meta):
        pass

    # FIXME not optimized at all...
    def rule(self, ftype):
        if not hasattr(self, '_rules'):
            self._rules = {}
        if not self._rules.has_key(ftype):
            if not ftype.endswith(settings.FORMATTER_FIELD_SUFFIX):
                ftype = ftype + settings.FORMATTER_FIELD_SUFFIX
            fmtr = getattr(self, ftype)
            self._rules[ftype] = (fmtr, len(self.fields.all()))
        return self._rules[ftype]

    def add_fields_to_queryset(self, queryset, using=DEFAULT_MODELTREE_ALIAS, **kwargs):
        modeltree = trees[using]

        # ensures that new joins that are performed use LEFT OUTER JOINs so
        # records do not get thrown out
        kwargs['outer_if_first'] = True

        # TODO test if column being added is nullable? see
        # django/db/models/sql/query.py#L818

        fields = self.fields.order_by('columnfield__order')
        aliases = []
        for f in fields:
            queryset = modeltree.add_joins(f.model, queryset, **kwargs)
            aliases.append((f.model._meta.db_table, f.field.column))
        return (queryset, aliases)

    def get_ordering_for_queryset(self, direction='asc', using=DEFAULT_MODELTREE_ALIAS):
        fields = self.fields.all()
        orders = []
        for f in fields:
            orders.append(f.order_string(direction, using))
        return orders

class ColumnField(ConceptField):
    concept = models.ForeignKey(Column, related_name='conceptfields')
    field = models.ForeignKey(Field, limit_choices_to={'is_public': True})

    class Meta(ConceptField.Meta):
        pass



from django.db import models
from django.db.models import Q
from django.conf import settings

class FieldManager(models.Manager):
    "Adds additional helper methods focused around access and permissions."
    use_for_related_fields = True

    def get_by_natural_key(self, app_name, model_name, field_name):
        return self.get_query_set().get(app_name=app_name, model_name=model_name,
            field_name=field_name)

    def _get_for_site(self):
        return Q(sites=None) | Q(sites__id__exact=settings.SITE_ID)

    def _public_for_auth_user(self, user):
        kwargs = {'is_public': True}
        groups = Q(group=None) | Q(group__in=user.groups.all())
        sites = self._get_for_site()

        return self.get_query_set().filter(sites, groups, **kwargs).distinct()

    def _public_for_anon_user(self):
        kwargs = {
            'group': None,
            'is_public': True
        }
        sites = self._get_for_site()

        return self.get_query_set().filter(sites, **kwargs).distinct()

    def public(self, user=None):
        "Returns all publically available fields given a user."
        if user and user.is_authenticated():
            return self._public_for_auth_user(user)
        return self._public_for_anon_user()



from avocado.concepts.models import *
from avocado.fields.models import *
from avocado.columns.models import *
from avocado.criteria.models import *
from avocado.store.models import *

import avocado.fields.translate
import avocado.columns.format
import avocado.criteria.viewset



from django.utils.functional import LazyObject
from django.conf import settings as default_settings

from avocado.conf import global_settings

class LazySettings(LazyObject):
    def _setup(self):
        self._wrapped = Settings(getattr(default_settings, 'AVOCADO_SETTINGS', {}))


class Settings(object):
    def __init__(self, settings_dict):
        for setting in dir(global_settings):
            if setting == setting.upper():
                setattr(self, setting, getattr(global_settings, setting))

        for k, v in settings_dict.items():
            if k == k.upper():
                setattr(self, k, v)

settings = LazySettings()



import math

from django.core.paginator import Paginator, Page

class BufferedPaginator(Paginator):
    """A subclass of the Django Paginator class that allows for explicitly
    setting the ``_count`` attribute i.e. the theoretical size of
    ``object_list``. This removes the need for the Paginator to do determine
    the size of the ``object_list`` which improves performance especially for
    complex queries (and removes the need to actually pass in the
    ``objec_list``).

    The ``object_list`` that is passed in as an argument should represent the
    *buffered* part of the whole ``object_list``. This works in conjunction
    with the ``BufferedPage`` which can test whether or not a particular page
    is actually stored in the ``BufferedPaginator`` object.

    If ``object_list`` is not passed in, all calculations will still be
    available, but any operations that act on the ``object_list`` will
    throw an error, unless an ``object_list`` is passed in.

    The use of the class is primarily for large data sets in which it is
    impractical to store the entire ``object_list`` into a paginator.
    Overriding ``_count`` allows for the methods to still be used.

    ``buf_size`` represents the size of the buffer, i.e. number of rows that will
    be available at any given time and in most cases the size of ``object_list``
    assuming ``_count`` is greater than ``buf_size``..
    """
    def __init__(self, count, object_list=None, offset=0, buf_size=None, *args, **kwargs):
        if not object_list and not buf_size:
            raise ValueError, 'an "object_list" or a "buf_size" must be defined'
        if offset > count:
            raise ValueError, '"offset" cannot be greater than the "count"'

        super(BufferedPaginator, self).__init__(object_list, *args, **kwargs)

        self._count = count
        # negative offsets are relative to to ``count``
        if offset < 0:
            offset = count + offset
        self.offset = offset
        self.buf_size = buf_size or len(object_list)

    def page(self, number):
        """Returns a ``BufferedPage`` object representing the slice of data
        relative to the ``per_page`` number.
        """
        number = self.validate_number(number)

        if self.object_list:
            bottom = (number - 1) * self.per_page - self.offset

            if bottom < 0:
                bottom = 0

            top = bottom + self.per_page

            if top + self.orphans >= self.count:
                top = self.count

            object_list = self.object_list[bottom:top]
        elif self.object_list is not None:
            object_list = []
        else:
            object_list = None

        return BufferedPage(number, self, object_list)

    def cached_page_indices(self):
        "Returns the indices of the first and last pages that are cached."
        # the number of rows per page is greater than we will ever have in
        # the buffer at any given time. we can only provide partial results
        if self.buf_size < self.per_page:
            # the beginning items are missing.. no good
            if self.offset > 0:
                return (0, 0)

        pages_of_offset = math.ceil(self.offset / float(self.per_page))
        first = int(pages_of_offset) + 1
        last = first + math.ceil(self.buf_size / float(self.per_page))

        return (int(first), int(last))

    def cached_pages(self):
        first, last = self.cached_page_indices()
        return last - first

    def get_overlap(self, offset, buf_size=None):
        """Calculates the overlap of the buffered rows relative to a new offset
        and (optionally) buf_size. There are two sets of outputs here including
        the start and end terminals::

            [(start_offset, start_limit), (end_offset, end_limit)]

        The offsets and limits can be used for list slicing or for SQL OFFSET/
        LIMIT clauses.
        """
        buf_size = buf_size or self.buf_size

        overlap = True
        start_offset = start_limit = end_offset = end_limit = None

        # if the start terminal is within the current range, then there will be
        # a slice of data to be appended
        if self.offset <= offset <= (self.offset + self.buf_size):
            end_offset = (self.offset + self.buf_size) + 1
            end_limit = buf_size - (end_offset - offset) + 1

        # if the end terminal is within the current range, then there will be a
        # slice of data to be prepended
        elif self.offset <= (offset + buf_size) <= (self.offset + self.buf_size):
            start_offset = offset
            start_limit = self.offset - start_offset

        # check to see if the current range is within the start and end
        # terminals
        elif offset <= self.offset and (self.offset + self.buf_size) <= (offset + buf_size):
            start_offset = offset
            start_limit = self.offset - start_offset
            end_offset = (self.offset + self.buf_size) + 1
            end_limit = (offset + buf_size) - end_offset + 1

        # no overlap at all
        else:
            overlap = False
            start_offset = offset
            start_limit = buf_size

        return (overlap, (start_offset, start_limit), (end_offset, end_limit))


class BufferedPage(Page):
    """A subclass of the django Page class which adds an additional method to
    determine if this page is in cache. The determination is with respect to
    the `paginator' `offset' and `buf_size' attributes.
    """
    def __init__(self, number, paginator, object_list=None):
        self.number = number
        self.paginator = paginator
        self.object_list = object_list

    def in_cache(self):
        first, last = self.paginator.cached_page_indices()
        return first <= self.number < last

    def offset(self):
        return max(self.start_index(), 1) - 1

    def get_list(self, object_list=None):
        if object_list is not None:
            s = self.start_index() - self.paginator.offset - 1
            e = self.end_index() - self.paginator.offset
            return object_list[s:e]
        if self.object_list is not None:
            return self.object_list
        return None

    def page_links(self, before_after_count=2, first_last_count=0):
        pages = []
        # only process if the current page number is greater than the number of
        # required links.
        if self.number > (first_last_count + before_after_count + 1):
            # append first pages, e.g. [1, 2] if `first_last_count' is 2
            for i in range(1, first_last_count + 1):
                pages.append(i)

            pages.append(None)
            # append the leading pages relative to the current page, e.g.
            # [10, 11] if `before_after_count' is 2 and current page is 12
            for i in range(self.number - before_after_count, self.number):
                pages.append(i)

        else:
            for i in range(1, self.number):
                pages.append(i)

        # if there is still a gap until `self.paginator.num_pages' is reached, add trailing pages
        if (self.number + first_last_count + before_after_count) < self.paginator.num_pages:
            for i in range(self.number, self.number + before_after_count + 1):
                pages.append(i)

            pages.append(None)
            for i in range(self.paginator.num_pages + 1 - first_last_count, self.paginator.num_pages + 1):
                pages.append(i)

        else:
            for i in range(self.number, self.paginator.num_pages + 1):
                pages.append(i)

        return pages



"App-level admin module so all subapps admins get registered."

from avocado.fields.admin import *
from avocado.columns.admin import *
from avocado.criteria.admin import *
from avocado.concepts.admin import *


import datetime
try:
    from cPickle import loads, dumps
except ImportError:
    from pickle import loads, dumps
from copy import deepcopy
from base64 import b64encode, b64decode
from zlib import compress, decompress

from django.conf import settings
from django.db import models
from django.db.models import signals
from django.utils import simplejson as json
from django.utils.translation import ugettext_lazy as _
from django.utils.encoding import force_unicode

class PickledObject(str):
    """
    A subclass of string so it can be told whether a string is a pickled
    object or not (if the object is an instance of this class then it must
    [well, should] be a pickled one).

    Only really useful for passing pre-encoded values to ``default``
    with ``dbsafe_encode``, not that doing so is necessary. If you
    remove PickledObject and its references, you won't be able to pass
    in pre-encoded values anymore, but you can always just pass in the
    python objects themselves.

    """
    pass

def dbsafe_encode(value, compress_object=False):
    """
    We use deepcopy() here to avoid a problem with cPickle, where dumps
    can generate different character streams for same lookup value if
    they are referenced differently.

    The reason this is important is because we do all of our lookups as
    simple string matches, thus the character streams must be the same
    for the lookups to work properly. See tests.py for more information.
    """
    if not compress_object:
        value = b64encode(dumps(deepcopy(value)))
    else:
        value = b64encode(compress(dumps(deepcopy(value))))
    return PickledObject(value)

def dbsafe_decode(value, compress_object=False):
    if not compress_object:
        value = loads(b64decode(value))
    else:
        value = loads(decompress(b64decode(value)))
    return value

class PickledField(models.Field):
    """
    A field that will accept *any* python object and store it in the
    database. PickledField will optionally compress it's values if
    declared with the keyword argument ``compress=True``.

    Does not actually encode and compress ``None`` objects (although you
    can still do lookups using None). This way, it is still possible to
    use the ``isnull`` lookup type correctly. Because of this, the field
    defaults to ``null=True``, as otherwise it wouldn't be able to store
    None values since they aren't pickled and encoded.

    """
    __metaclass__ = models.SubfieldBase

    def __init__(self, *args, **kwargs):
        self.compress = kwargs.pop('compress', False)
        self.protocol = kwargs.pop('protocol', 2)
        kwargs.setdefault('null', True)
        kwargs.setdefault('editable', False)
        super(PickledField, self).__init__(*args, **kwargs)

    def get_default(self):
        """
        Returns the default value for this field.

        The default implementation on models.Field calls force_unicode
        on the default, which means you can't set arbitrary Python
        objects as the default. To fix this, we just return the value
        without calling force_unicode on it. Note that if you set a
        callable as a default, the field will still call it. It will
        *not* try to pickle and encode it.

        """
        if self.has_default():
            if callable(self.default):
                return self.default()
            return self.default
        # If the field doesn't have a default, then we punt to models.Field.
        return super(PickledField, self).get_default()

    def to_python(self, value):
        """
        B64decode and unpickle the object, optionally decompressing it.

        If an error is raised in de-pickling and we're sure the value is
        a definite pickle, the error is allowed to propogate. If we
        aren't sure if the value is a pickle or not, then we catch the
        error and return the original value instead.

        """
        if value is not None:
            try:
                value = dbsafe_decode(value, self.compress)
            except:
                # If the value is a definite pickle; and an error is raised in
                # de-pickling it should be allowed to propogate.
                if isinstance(value, PickledObject):
                    raise
        return value

    def get_db_prep_value(self, value):
        """
        Pickle and b64encode the object, optionally compressing it.

        The pickling protocol is specified explicitly (by default 2),
        rather than as -1 or HIGHEST_PROTOCOL, because we don't want the
        protocol to change over time. If it did, ``exact`` and ``in``
        lookups would likely fail, since pickle would now be generating
        a different string.

        """
        if value is not None and not isinstance(value, PickledObject):
            # We call force_unicode here explicitly, so that the encoded string
            # isn't rejected by the postgresql_psycopg2 backend. Alternatively,
            # we could have just registered PickledObject with the psycopg
            # marshaller (telling it to store it like it would a string), but
            # since both of these methods result in the same value being stored,
            # doing things this way is much easier.
            value = force_unicode(dbsafe_encode(value, self.compress))
        return value

    def value_to_string(self, obj):
        value = self._get_val_from_obj(obj)
        return self.get_db_prep_value(value)

    def get_internal_type(self):
        return 'TextField'

    def get_db_prep_lookup(self, lookup_type, value):
        if lookup_type not in ['exact', 'in', 'isnull']:
            raise TypeError('Lookup type %s is not supported.' % lookup_type)
        # The Field model already calls get_db_prep_value before doing the
        # actual lookup, so all we need to do is limit the lookup types.
        return super(PickledField, self).get_db_prep_lookup(lookup_type, value)


class JSONDateEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.strftime('%Y-%m-%d %H:%M:%S')
        elif isinstance(obj, datetime.date):
            return obj.strftime('%Y-%m-%d')
        elif isinstance(obj, datetime.time):
            return obj.strftime('%H:%M:%S')
        return json.JSONEncoder.default(self, obj)


class JSONField(models.TextField):
    description = _("Data that serializes and deserializes into and out of JSON.")

    def _dumps(self, data):
        return JSONDateEncoder().encode(data)

    def _loads(self, string):
        return json.loads(string, encoding=settings.DEFAULT_CHARSET)

    def pre_save(self, model_instance, add):
        value = getattr(model_instance, self.attname, None)
        return self._dumps(value)

    def contribute_to_class(self, cls, name):
        self.class_name = cls
        super(JSONField, self).contribute_to_class(cls, name)
        signals.post_init.connect(self.post_init)

        def get_json(model_instance):
            return self._dumps(getattr(model_instance, self.attname, None))
        setattr(cls, 'get_%s_json' % self.name, get_json)

        def set_json(model_instance, json):
            return setattr(model_instance, self.attname, self._loads(json))
        setattr(cls, 'set_%s_json' % self.name, set_json)

    def post_init(self, **kwargs):
        if 'sender' in kwargs and 'instance' in kwargs:
            if kwargs['sender'] == self.class_name and hasattr(kwargs['instance'], self.attname):
                value = self.value_from_object(kwargs['instance'])
                if (value):
                    setattr(kwargs['instance'], self.attname, self._loads(value))
                else:
                    setattr(kwargs['instance'], self.attname, None)


# support for South
try:
    from south.modelsinspector import add_introspection_rules
    add_introspection_rules([], ['^avocado\.store\.fields\.JSONField'])
    add_introspection_rules([
        (
            (PickledField,),
            [],
            {
                'compress': ['compress', {'default': False}],
                'protocol': ['protocol', {'default': 2}],
                'null': ['null', {'default': True}],
                'editable': ['editable', {'editable': False}]
            }
        )
    ], ['^avocado\.store\.fields'])

except ImportError:
    pass



from django.core.management.base import NoArgsCommand

from avocado.models import Field

class Command(NoArgsCommand):
    help = "Determines outdated or orphaned Field definitions."

    def handle_noargs(self, **options):
        fields = Field.objects.all()

        unknown_model = []
        unknown_field = []

        for f in fields:
            if f.model is None:
                unknown_model.append(f)
            elif f.field is None:
                unknown_field.append(f)

        if unknown_model:
            print
            print 'The following Fields have an unknown model:\n'
            for f in unknown_model:
                print '\t',
                if f.is_public:
                    print '[A]',
                else:
                    print '   ',
                print f
        if unknown_field:
            print
            print 'The following Fields have an unknown field:\n'
            for f in unknown_field:
                print '\t',
                if f.is_public:
                    print '[A]',
                else:
                    print '   ',
                print f
            print



from django.core.exceptions import ObjectDoesNotExist
from django.core.cache import cache

class ConceptCache(object):
    def __init__(self, model_name):
        self.id_key = ':'.join(['avocado', model_name, '%s'])
        self.field_id_key = ':'.join(['avocado', model_name, '%s', 'fields'])

    def get(self, concept_id, queryset=None, ret_val=None):
        "Simple interface for getting (and setting) a concept from global cache."
        key = self.id_key % concept_id

        concept = cache.get(key, None)
        if concept is None:
            try:
                concept = queryset.get(id=concept_id)
            except ObjectDoesNotExist:
                return ret_val
            cache.set(key, concept)
        return concept

    def get_many(self, concept_ids, queryset=None):
        "Returns a generator of concept objects."
        concepts = []
        for cid in concept_ids:
            concepts.append(self.get(cid, queryset=queryset))
        return concepts


    def get_fields(self, concept_id, queryset, ret_val=None):
        """Simple interface for getting (and setting) a concept's fields
        from global cache. `concept' can be an object or a integer.

        The optional `queryset' parameter simply passes it to `self.get'
        if called.
        """
        key = self.field_id_key % concept_id

        fields = cache.get(key, None)
        if fields is None:
            concept = self.get(concept_id, queryset, ret_val)
            if concept == ret_val:
                return ret_val
            fields = list(concept.fields.order_by('conceptfield__order'))
            cache.set(key, fields)
        return fields



import re

class CamelCaser(object):
    UNDERSCORE = re.compile(r'([^\A_])_+([^_])')

    def __call__(self, s):
        return self.camel(s)

    def _upper(self, m):
        f, l = m.groups()
        return '%s%s' % (f, l.upper())

    def camel(self, s):
        return self.UNDERSCORE.sub(self._upper, s)

    def camel_keys(self, rdict, cdict=None):
        cdict = cdict or {}
        for k, v in rdict.items():
            ck = self.camel(k)
            if type(v) is dict:
                cdict[ck] = self.camel_keys(v)
            else:
                cdict[ck] = v
        return cdict


class UnCamelCaser(object):
    def __call__(self, s):
        return self.uncamel(s)

    def uncamel(self, s):
        toks = [s[0]]
        for x in s[1:]:
            if x.isupper():
                toks.append(' ')
            toks.append(x)
        return ''.join(toks)


camel= CamelCaser()
uncamel = UnCamelCaser()



from avocado.concepts.cache import ConceptCache
from avocado.criteria.models import Criterion

class CriterionCache(ConceptCache):
    def get(self, concept_id, queryset=None, ret_val=None):
        if queryset is None:
            queryset = Criterion.objects.all()
        return super(CriterionCache, self).get(concept_id=concept_id,
            queryset=queryset, ret_val=ret_val)
    
    def get_fields(self, concept_id, queryset=None, ret_val=None):
        if queryset is None:
            queryset = Criterion.objects.all()
        return super(CriterionCache, self).get_fields(concept_id=concept_id,
            queryset=queryset, ret_val=ret_val)


cache = CriterionCache(Criterion.__name__.lower())


def postgresql_psycopg2(queryset, terms):
    column = '"%s"."search_tsv"' % queryset.model._meta.db_table

    if not terms:
        return queryset

    queryset = queryset.extra(where=(column + ' @@ to_tsquery(%s)',),
        params=('&'.join(terms),))

    return queryset



from django.db.models import ObjectDoesNotExist

def objectset_scope_sync_pre_save(instance, **kwargs):
    """Pre-save handler for both the ``Scope`` object and the ``ObjectSet``
    object. If registered with ``Scope`` as the sender, two additional keyword
    arguments must be partially applied prior to connecting the handler.

    ``set_name`` - the reverse name of the ``ObjectSet`` subclass' instance

    ``set_manager`` - the name of the many-to-many manager on the relation

    If the sender is a subclass of ``ObjectSet``, only the ``set_manager``
    must be defined.
    """
    scope = instance
    try:
        objectset = getattr(scope, kwargs['set_name'])
    except ObjectDoesNotExist:
        return

    objectset_manager = getattr(objectset, kwargs['set_manager'])
    oids = set(objectset_manager.distinct().values_list('id', flat=True))
    nids = set(scope.get_queryset().distinct().values_list('id', flat=True))

    added = (oids | nids) - oids
    removed = (oids | nids) - nids

    if added:
        objectset_manager.add(*added)
    if removed:
        objectset_manager.remove(*removed)

def objectset_incr_count_pre_save(instance, action, pk_set, **kwargs):
    """Handlers syncing `count` based on the number of objects are being
    added or removed from the m2m relationship.
    """
    if action in ('pre_add', 'pre_remove', 'pre_clear'):
        if action == 'pre_clear':
            count = -instance.count
        else:
            count = len(pk_set)
            if action == 'pre_remove':
                count = -count

        instance.count += count
        instance.save()





from django.db import models
from django.db.models import Q
from django.template import Template, Context
from avocado.concepts.managers import ConceptManager

__all__ = ('Category',)

class Category(models.Model):
    name = models.CharField(max_length=100)
    parent = models.ForeignKey('self', null=True, blank=True,
            related_name='children', limit_choices_to={'parent__isnull': True})
    order = models.FloatField(default=0)

    class Meta(object):
        verbose_name_plural = 'categories'
        app_label = 'avocado'
        ordering = ('order', 'name')

    def __unicode__(self):
        return u'%s' % self.name


class Concept(models.Model):
    # categories that do not have children or have a parent
    category_lookup = Q(children__isnull=True) | Q(parent__isnull=False)

    name = models.CharField(max_length=100)
    description = models.TextField(null=True, blank=True)
    keywords = models.CharField(max_length=100, null=True, blank=True)
    category = models.ForeignKey(Category, null=True, blank=True,
        limit_choices_to=category_lookup)
    is_public = models.BooleanField(default=False)
    order = models.FloatField(default=0, help_text='This ' \
        'ordering is relative to the category this concept belongs to.')

    # search optimizations
    search_doc = models.TextField(editable=False, null=True)

    objects = ConceptManager()

    class Meta(object):
        abstract = True
        app_label = 'avocado'
        ordering = ('name',)

    def __unicode__(self):
        return u'%s' % self.name

    def full_description(self):
        if self.description:
            fields = [{'name': self.name, 'description': self.description}]
        else:
            fields = list(self.conceptfields.select_related('field').values('name',
                'field__name', 'field__description').exclude(field__description=None))

            for x in fields:
                if x['name'] is None:
                    x['name'] = x.pop('field__name')
                x['description'] = x.pop('field__description').strip()

            # edge case when pretty field names are not set
            if len(fields) == 1:
                fields[0]['name'] = self.name

        c = Context({'fields': fields})
        t = Template("""
            {% load markup %}
            {% spaceless %}
            {% for field in fields %}
                {{ field.description|markdown }}
            {% endfor %}
            {% endspaceless %}
            """)
        return t.render(c).strip()


class ConceptField(models.Model):
    name = models.CharField(max_length=100, null=True, blank=True)
    order = models.FloatField(default=0)

    class Meta(object):
        abstract = True
        app_label = 'avocado'
        ordering = ('order',)

    def __unicode__(self):
        return u'%s' % self.field
